# mypy: disable-error-code="no-untyped-def"

import warnings
from typing import Any

import pandas as pd
import pytest
import regex

from dsp_tools.error.exceptions import InputError
from dsp_tools.error.xmllib_warnings import XmllibInputInfo
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib import LicenseRecommended
from dsp_tools.xmllib.internal.input_converters import check_and_fix_collection_input
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.internal.file_values import FileValue
from dsp_tools.xmllib.models.internal.file_values import IIIFUri
from dsp_tools.xmllib.models.internal.values import BooleanValue
from dsp_tools.xmllib.models.internal.values import ColorValue
from dsp_tools.xmllib.models.internal.values import DateValue
from dsp_tools.xmllib.models.internal.values import DecimalValue
from dsp_tools.xmllib.models.internal.values import GeonameValue
from dsp_tools.xmllib.models.internal.values import IntValue
from dsp_tools.xmllib.models.internal.values import LinkValue
from dsp_tools.xmllib.models.internal.values import ListValue
from dsp_tools.xmllib.models.internal.values import Richtext
from dsp_tools.xmllib.models.internal.values import SimpleText
from dsp_tools.xmllib.models.internal.values import TimeValue
from dsp_tools.xmllib.models.internal.values import UriValue
from dsp_tools.xmllib.models.res import Resource


class TestCreateNewResource:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            Resource.create_new("res_id", "restype", "label")
            assert len(caught_warnings) == 0

    def test_empty_resource_id(self):
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid xsd:ID, your input '' does not match the type."),
        ):
            Resource.create_new("", "restype", "label")

    def test_invalid_resource_id(self):
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid xsd:ID, your input 'not|ok' does not match the type."),
        ):
            Resource.create_new("not|ok", "restype", "label")

    def test_empty_restype(self):
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid resource type, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "", "label")

    def test_empty_label(self):
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid non empty string, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "")


class TestAddValues:
    def test_add_bool(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_bool(":prop", "True")
        assert len(res.values) == 1
        assert isinstance(res.values[0], BooleanValue)

    def test_add_bool_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid bool, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_bool("", "")

    def test_add_bool_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_bool_optional("", None)
        assert not res.values
        res = res.add_bool_optional(":prop", "True")
        assert len(res.values) == 1
        assert isinstance(res.values[0], BooleanValue)

    def test_add_color(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_color(":prop", "#000000")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ColorValue)

    def test_add_color_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid color, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_color("", "")

    def test_add_color_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_color_multiple(":prop", ["#000000", "#ffffff"])
        assert len(res.values) == 2
        assert all([isinstance(x, ColorValue) for x in res.values])

    def test_add_color_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_color_optional("", pd.NA)
        assert not res.values
        res = res.add_color_optional(":prop", "#000000")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ColorValue)

    def test_add_date(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_date(":prop", "2024-10-31")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DateValue)

    def test_add_date_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid date, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_date("", "")

    def test_add_date_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_date_multiple(":prop", ["2024-10-30", "2024-10-31"])
        assert len(res.values) == 2
        assert all([isinstance(x, DateValue) for x in res.values])

    def test_add_date_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_date_optional("", None)
        assert not res.values
        res = res.add_date_optional(":prop", "2024-10-30")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DateValue)

    def test_add_decimal(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_decimal(":prop", "0.1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DecimalValue)

    def test_add_decimal_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid decimal, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_decimal("", "")

    def test_add_decimal_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_decimal_multiple(":prop", ["0.1", "0.2"])
        assert len(res.values) == 2
        assert all([isinstance(x, DecimalValue) for x in res.values])

    def test_add_decimal_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_decimal_optional("", None)
        assert not res.values
        res = res.add_decimal_optional(":prop", "0.1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DecimalValue)

    def test_add_geoname(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_geoname(":prop", "123456")
        assert len(res.values) == 1
        assert isinstance(res.values[0], GeonameValue)

    def test_add_geoname_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid geoname, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_geoname("", "")

    def test_add_geoname_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_geoname_multiple(":prop", ["123456", "7890"])
        assert len(res.values) == 2
        assert all([isinstance(x, GeonameValue) for x in res.values])

    def test_add_geoname_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_geoname_optional("", pd.NA)
        assert not res.values
        res = res.add_geoname_optional(":prop", "123456")
        assert len(res.values) == 1
        assert isinstance(res.values[0], GeonameValue)

    def test_add_integer(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_integer(":prop", "1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], IntValue)

    def test_add_integer_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid integer, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_integer("", "")

    def test_add_integer_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_integer_multiple(":prop", ["1", "2"])
        assert len(res.values) == 2
        assert all([isinstance(x, IntValue) for x in res.values])

    def test_add_integer_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_integer_optional("", None)
        assert not res.values
        res = res.add_integer_optional(":prop", "1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], IntValue)

    def test_add_link(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_link(":prop", "other_id")
        assert len(res.values) == 1
        assert isinstance(res.values[0], LinkValue)

    def test_add_link_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape(
                "The input should be a valid xsd:ID or DSP resource IRI, your input '' does not match the type."
            ),
        ):
            Resource.create_new("res_id", "restype", "label").add_link("", "")

    def test_add_link_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_link_multiple(
            ":prop", ["other_id", "yet_another_id"]
        )
        assert len(res.values) == 2
        assert all([isinstance(x, LinkValue) for x in res.values])

    def test_add_link_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_link_optional("", pd.NA)
        assert not res.values
        res = res.add_link_optional(":prop", "other_id")
        assert len(res.values) == 1
        assert isinstance(res.values[0], LinkValue)

    def test_add_list(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_list(":prop", "list-1", "node-1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ListValue)

    def test_add_list_node_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid list node, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_list("", "listname", "")

    def test_add_list_name_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid list name, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_list("", "", "node")

    def test_add_list_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_list_multiple(
            ":prop", "list-1", ["node-1", "node-2"]
        )
        assert len(res.values) == 2
        assert all([isinstance(x, ListValue) for x in res.values])

    def test_add_list_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_list_optional(":prop", "listname", None)
        assert not res.values
        res = res.add_list_optional(":prop", "list-1", "node-1")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ListValue)

    def test_add_simple_text(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_simpletext(":prop", "text")
        assert len(res.values) == 1
        assert isinstance(res.values[0], SimpleText)

    def test_add_simple_text_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("Your input '' is empty. Please enter a valid string."),
        ):
            Resource.create_new("res_id", "restype", "label").add_simpletext("", "")

    def test_add_simple_text_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_simpletext_multiple(":prop", ["text1", "text2"])
        assert len(res.values) == 2
        assert all([isinstance(x, SimpleText) for x in res.values])

    def test_add_simple_text_multiple_input_non_list(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_simpletext_multiple(":prop", "text1")
        assert len(res.values) == 1
        assert all([isinstance(x, SimpleText) for x in res.values])
        assert res.values[0].value == "text1"

    def test_add_simple_text_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_simpletext_optional(":prop", None)
        assert not res.values
        res = res.add_simpletext_optional(":prop", "text")
        assert len(res.values) == 1
        assert isinstance(res.values[0], SimpleText)

    def test_add_richtext(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_richtext(":prop", "A\nB")
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)
        assert res.values[0].value == "A<br/>B"

    def test_add_richtext_warns_empty_string(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("Your input '' is empty. Please enter a valid string."),
        ):
            Resource.create_new("res_id", "restype", "label").add_richtext("prop", "")

    def test_add_richtext_warns_pd_na(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("Your input '<NA>' is empty. Please enter a valid string."),
        ):
            Resource.create_new("res_id", "restype", "label").add_richtext(":prop", pd.NA)  # type: ignore[arg-type]

    def test_add_richtext_no_replace(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_richtext(
            "", "A\nB", newline_replacement=NewlineReplacement.NONE
        )
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)
        assert res.values[0].value == "A\nB"

    def test_add_richtext_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_richtext_multiple(":prop", ["text1", "text2"])
        assert len(res.values) == 2
        assert all([isinstance(x, Richtext) for x in res.values])

    def test_add_richtext_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_richtext_optional(":prop", None)
        assert not res.values
        res = res.add_richtext_optional(":prop", "text")
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)

    def test_add_time(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_time(":prop", "2024-10-31T11:00:00Z")
        assert len(res.values) == 1
        assert isinstance(res.values[0], TimeValue)

    def test_add_time_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid timestamp, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_time("", "")

    def test_add_time_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_time_multiple(
            ":prop", ["2024-10-31T11:00:00Z", "2023-01-01T00:00:00Z"]
        )
        assert len(res.values) == 2
        assert all([isinstance(x, TimeValue) for x in res.values])

    def test_add_time_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_time_optional(":prop", None)
        assert not res.values
        res = res.add_time_optional(":prop", "2023-01-01T00:00:00Z")
        assert len(res.values) == 1
        assert isinstance(res.values[0], TimeValue)

    def test_add_uri(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_uri(":prop", "https://www.google.com/")
        assert len(res.values) == 1
        assert isinstance(res.values[0], UriValue)

    def test_add_uri_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid uri, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_uri("", "")

    def test_add_uri_multiple(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_uri_multiple(
            "", ["https://www.google.com/", "https://dasch.swiss/"]
        )
        assert len(res.values) == 2
        assert all([isinstance(x, UriValue) for x in res.values])

    def test_add_uri_optional(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_uri_optional(":prop", pd.NA)
        assert not res.values
        res = res.add_uri_optional(":prop", "https://dasch.swiss/")
        assert len(res.values) == 1
        assert isinstance(res.values[0], UriValue)


class TestAddFiles:
    def test_add_file(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_file(
            "foo/bar.baz", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"]
        )
        assert isinstance(res.file_value, FileValue)

    def test_add_file_with_metadata(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_file(
            "foo/bar.baz", LicenseRecommended.CC.BY, "copy", ["auth"]
        )
        assert isinstance(res.file_value, FileValue)
        assert res.file_value.metadata.license == LicenseRecommended.CC.BY
        assert res.file_value.metadata.copyright_holder == "copy"
        assert res.file_value.metadata.authorship == tuple(["auth"])

    def test_add_file_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("Field 'bitstream' | Your input '' is empty. Please enter a valid file name."),
        ):
            Resource.create_new("res_id", "restype", "label").add_file(
                "", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"]
            )

    def test_add_file_raising(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_file(
            "existing filename", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"]
        )
        msg = regex.escape(
            "Resource ID 'res_id' | Field 'file / iiif-uri' | "
            "This resource already contains a file with the name: 'existing filename'. "
            "The new file with the name 'new filename' cannot be added."
        )
        with pytest.raises(InputError, match=msg):
            res.add_file("new filename", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"])

    def test_add_iiif_uri(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_iiif_uri(
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2",
            LicenseRecommended.DSP.UNKNOWN,
            "copy",
            ["auth"],
        )
        assert isinstance(res.file_value, IIIFUri)

    def test_add_iiif_uri_with_metadata(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_iiif_uri(
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2",
            LicenseRecommended.CC.BY,
            "copy",
            ["auth"],
        )
        assert isinstance(res.file_value, IIIFUri)
        assert res.file_value.metadata.license == LicenseRecommended.CC.BY
        assert res.file_value.metadata.copyright_holder == "copy"
        assert res.file_value.metadata.authorship == tuple(["auth"])

    def test_add_iiif_uri_warns(self) -> None:
        with pytest.warns(
            XmllibInputWarning,
            match=regex.escape("The input should be a valid IIIF uri, your input '' does not match the type."),
        ):
            Resource.create_new("res_id", "restype", "label").add_iiif_uri(
                "", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"]
            )

    def test_add_iiif_uri_raising(self) -> None:
        res = Resource.create_new("res_id", "restype", "label").add_file(
            "existing IIIF", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"]
        )
        msg = regex.escape(
            "Resource ID 'res_id' | Field 'file / iiif-uri' | "
            "This resource already contains a file with the name: 'existing IIIF'. "
            "The new file with the name 'new IIIF' cannot be added."
        )
        with pytest.raises(InputError, match=msg):
            res.add_iiif_uri("new IIIF", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth"])


@pytest.mark.parametrize(
    ("input_val", "expected_val"),
    [(1, [1]), (True, [True]), ("string", ["string"]), ([1, 2], [1, 2]), ((1, 2), [1, 2]), ({1, 2}, [1, 2])],
)
def test_check_and_fix_collection_input_success(input_val: Any, expected_val: list[Any]) -> None:
    assert check_and_fix_collection_input(input_val, "prop", "id") == sorted(expected_val)


def test_check_and_fix_collection_input_warns() -> None:
    msg = regex.escape(
        "Resource ID 'id' | Property 'prop' | The input is empty. "
        "Please note that no values will be added to the resource."
    )
    with pytest.warns(XmllibInputInfo, match=msg):
        check_and_fix_collection_input([], "prop", "id")


def test_check_and_fix_collection_input_raises() -> None:
    msg = regex.escape(
        "Resource ID 'id' | Property 'prop' | "
        "The input is a dictionary. Only collections (list, set, tuple) are permissible."
    )
    with pytest.raises(InputError, match=msg):
        check_and_fix_collection_input({1: 1}, "prop", "id")


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest
import regex

from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_string
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import assert_is_tuple
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_boolean
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_date
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_decimal
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_geometry
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_integer
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_richtext
from dsp_tools.commands.xmlupload.prepare_xml_input.transform_input_values import transform_simpletext
from dsp_tools.error.exceptions import XmlInputConversionError
from dsp_tools.utils.data_formats.date_util import Calendar
from dsp_tools.utils.data_formats.date_util import Date
from dsp_tools.utils.data_formats.date_util import Era
from dsp_tools.utils.data_formats.date_util import SingleDate


def test_assert_is_string_good():
    result = assert_is_string("string")
    assert isinstance(result, str)
    assert result == "string"


def test_assert_is_string_raises():
    expected = r"Expected string value, but got tuple value: ('1', '2')"
    with pytest.raises(XmlInputConversionError, match=regex.escape(expected)):
        assert_is_string(("1", "2"))


def test_assert_is_tuple_good():
    result = assert_is_tuple(("1", "2"))
    assert isinstance(result, tuple)
    assert result == ("1", "2")


@pytest.mark.parametrize(
    ("in_val", "expected"),
    [
        (("1", "2", "3"), r"Expected tuple with two elements but got ('1', '2', '3')"),
        ("string", r"Expected tuple value, but got string value: string"),
    ],
)
def test_assert_is_tuple_raises(in_val, expected):
    with pytest.raises(XmlInputConversionError, match=regex.escape(expected)):
        assert_is_tuple(in_val)


@pytest.mark.parametrize(
    ("in_val", "expected"),
    [
        ("true", True),
        ("1", True),
        ("True", True),
        (1, True),
        ("false", False),
        ("False", False),
        ("0", False),
        (0, False),
    ],
)
def test_transform_boolean_good(in_val, expected):
    assert transform_boolean(in_val) == expected


def test_transform_boolean_raises():
    msg = regex.escape(r"Could not parse boolean value: other")
    with pytest.raises(XmlInputConversionError, match=msg):
        transform_boolean("other")


def test_transform_date_good():
    result = transform_date("2022")
    assert isinstance(result, Date)
    assert result.calendar == Calendar.GREGORIAN
    assert result.start.era == Era.CE
    assert result.start.year == 2022
    assert not result.start.month
    assert not result.start.day
    end_date = result.end
    assert isinstance(end_date, SingleDate)
    assert end_date.era == Era.CE
    assert end_date.year == 2022
    assert not end_date.month
    assert not end_date.day


@pytest.mark.parametrize(
    ("in_val", "expected"),
    [
        ("1", 1.0),
        ("1.88", 1.88),
    ],
)
def test_transform_decimal_good(in_val, expected):
    assert transform_decimal(in_val) == expected


def test_transform_integer_good():
    assert transform_integer("1") == 1


def test_transform_geometry_good():
    geom_str = """
                {
                    "status": "active",
                    "type": "polygon",
                    "lineWidth": 5,
                    "points": [{"x": 0.4, "y": 0.6},
                               {"x": 0.5, "y": 0.9},
                               {"x": 0.8, "y": 0.9},
                               {"x": 0.7, "y": 0.6}]
                }
                """
    expected = (
        '{"status": "active", "type": "polygon", "lineWidth": 5, "points": '
        '[{"x": 0.4, "y": 0.6}, {"x": 0.5, "y": 0.9}, {"x": 0.8, "y": 0.9}, {"x": 0.7, "y": 0.6}]}'
    )
    assert transform_geometry(geom_str) == expected


def test_transform_geometry_raises():
    msg = regex.escape(r"Could not parse json value: not valid")
    with pytest.raises(XmlInputConversionError, match=msg):
        transform_geometry("not valid")


def test_transform_simpletext_good():
    assert transform_simpletext("text") == "text"


def test_transform_simpletext_raises():
    msg = regex.escape(r"After removing redundant whitespaces and newlines the input string is empty.")
    with pytest.raises(XmlInputConversionError, match=msg):
        transform_simpletext("")


def test_transform_richtext_good():
    expected = "This is <em>italicized and <strong>bold</strong></em> text!"
    result = transform_richtext(expected)
    assert isinstance(result, FormattedTextValue)
    assert result.xmlstr == expected


def test_transform_richtext_raises():
    msg = regex.escape(r"After removing redundant whitespaces and newlines the input string is empty.")
    with pytest.raises(XmlInputConversionError, match=msg):
        transform_richtext("")



# mypy: disable-error-code="method-assign,no-untyped-def"

from uuid import uuid4

import pytest
from rdflib import RDF
from rdflib import XSD
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.formatted_text_value import FormattedTextValue
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedRichtext
from dsp_tools.commands.xmlupload.stash.stash_models import StandoffStashItem
from dsp_tools.commands.xmlupload.stash.upload_stashed_xml_texts import _make_richtext_update_graph
from dsp_tools.commands.xmlupload.stash.upload_stashed_xml_texts import _serialise_richtext_for_update
from dsp_tools.utils.rdflib_constants import KNORA_API

ONTO_STR = "http://0.0.0.0:3333/ontology/9999/onto/v2#"

ONTO = Namespace(ONTO_STR)

RES_IRI_STR = "http://rdfh.ch/9999/res_one"
RES_IRI = URIRef(RES_IRI_STR)
RES_TYPE = ONTO.Resource

PROP_IRI = ONTO.hasText
VAL_IRI_STR = "http://rdfh.ch/9999/res_one/values/richtext"
VAL_IRI = URIRef(VAL_IRI_STR)


@pytest.fixture
def standoff_stash_item() -> StandoffStashItem:
    val = ProcessedRichtext(
        value=FormattedTextValue("text"),
        prop_iri=str(PROP_IRI),
        resource_references=set(),
        value_uuid=str(uuid4()),
        comment=None,
        permissions=None,
    )
    return StandoffStashItem("res_one", str(RES_TYPE), val)


@pytest.fixture
def iri_resolver() -> IriResolver:
    return IriResolver({"res_one": RES_IRI_STR})


def test_make_richtext_update_graph(standoff_stash_item, iri_resolver):
    result = _make_richtext_update_graph(standoff_stash_item, VAL_IRI_STR, RES_IRI_STR, iri_resolver)
    assert len(result) == 5
    assert next(result.objects(RES_IRI, RDF.type)) == RES_TYPE
    assert next(result.objects(RES_IRI, PROP_IRI)) == VAL_IRI
    value = next(result.objects(VAL_IRI, KNORA_API.textValueAsXml))
    assert value == Literal('<?xml version="1.0" encoding="UTF-8"?>\n<text>text</text>', datatype=XSD.string)
    mapping = next(result.objects(VAL_IRI, KNORA_API.textValueHasMapping))
    assert mapping == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")
    assert next(result.objects(VAL_IRI, RDF.type)) == KNORA_API.TextValue


def test_serialise_richtext_for_update(standoff_stash_item, iri_resolver):
    result = _serialise_richtext_for_update(standoff_stash_item, VAL_IRI_STR, RES_IRI_STR, iri_resolver)
    expected = {
        "@id": "http://rdfh.ch/9999/res_one",
        "@type": "http://0.0.0.0:3333/ontology/9999/onto/v2#Resource",
        "http://0.0.0.0:3333/ontology/9999/onto/v2#hasText": {
            "@id": "http://rdfh.ch/9999/res_one/values/richtext",
            "@type": "http://api.knora.org/ontology/knora-api/v2#TextValue",
            "http://api.knora.org/ontology/knora-api/v2#textValueAsXml": {
                "@type": "http://www.w3.org/2001/XMLSchema#string",
                "@value": '<?xml version="1.0" encoding="UTF-8"?>\n<text>text</text>',
            },
            "http://api.knora.org/ontology/knora-api/v2#textValueHasMapping": {
                "@id": "http://rdfh.ch/standoff/mappings/StandardMapping"
            },
        },
    }
    assert result == expected


def test_find_ids_referenced_in_salsah_links_one_link() -> None:
    one_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_set = one_link_KnoraStandoffXml.find_internal_ids()
    assert returned_set == {"r2_id"}


def test_find_ids_referenced_in_salsah_links_three_links() -> None:
    three_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>This is normal text'
            '<a class="salsah-link" href="IRI:r3_id:IRI">r3_id</a>'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_set = three_link_KnoraStandoffXml.find_internal_ids()
    assert returned_set == {"r2_id", "r3_id"}


def test__replace_internal_ids_with_iris_one_link() -> None:
    resolver = IriResolver({"r1_id": "r1_iri", "r2_id": "r2_iri", "r3_id": "r3_iri"})
    one_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_instance = one_link_KnoraStandoffXml.with_iris(resolver)
    expected_str = (
        '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
        '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
        '<a class="salsah-link" href="r2_iri">r2_id</a>'
        "</text></text-prop></resource>"
    )
    assert returned_instance.xmlstr == expected_str


def test__replace_internal_ids_with_iris_three_links() -> None:
    resolver = IriResolver({"r1_id": "r1_iri", "r2_id": "r2_iri", "r3_id": "r3_iri"})
    three_link_KnoraStandoffXml = FormattedTextValue(
        xmlstr=(
            '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
            '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>This is normal text'
            '<a class="salsah-link" href="IRI:r3_id:IRI">r3_id</a>'
            '<a class="salsah-link" href="IRI:r2_id:IRI">r2_id</a>'
            "</text></text-prop></resource>"
        )
    )
    returned_instance = three_link_KnoraStandoffXml.with_iris(resolver)
    expected_str = (
        '<resource label="r1_label" restype="r1_restype" id="r1_id" permissions="open">'
        '<text-prop name=":hasRichtext"><text permissions="open" encoding="xml">'
        '<a class="salsah-link" href="r2_iri">r2_id</a>This is normal text'
        '<a class="salsah-link" href="r3_iri">r3_id</a>'
        '<a class="salsah-link" href="r2_iri">r2_id</a>'
        "</text></text-prop></resource>"
    )
    assert returned_instance.xmlstr == expected_str


if __name__ == "__main__":
    pytest.main([__file__])



# mypy: disable-error-code="method-assign,no-untyped-def"
import pandas as pd
import pytest
import regex

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings import XmllibInputInfo
from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.error.xmllib_warnings_util import _filter_stack_frames
from dsp_tools.error.xmllib_warnings_util import _get_calling_code_context
from dsp_tools.error.xmllib_warnings_util import _get_stack_frame_number
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_info
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_type_mismatch_warning
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_warning
from dsp_tools.error.xmllib_warnings_util import get_user_message_string


@pytest.fixture
def message_info():
    return MessageInfo("msg", "id")


def test_emit_xmllib_input_info(message_info):
    expected = regex.escape("Resource ID 'id' | msg")
    with pytest.warns(XmllibInputInfo, match=expected):
        emit_xmllib_input_info(message_info)


def test_emit_xmllib_input_warning(message_info):
    expected = regex.escape("Resource ID 'id' | msg")
    with pytest.warns(XmllibInputWarning, match=expected):
        emit_xmllib_input_warning(message_info)


def test_emit_xmllib_input_type_mismatch_warning():
    expected = regex.escape(
        "' | Resource ID 'id' | Field 'field' | "
        "The input should be a valid string, your input '<NA>' does not match the type."
    )
    with pytest.warns(XmllibInputWarning, match=expected):
        emit_xmllib_input_type_mismatch_warning(expected_type="string", value=pd.NA, res_id="id", value_field="field")


class TestGetMessageString:
    def test_with_property(self):
        msg_info = MessageInfo("msg", "id", "prop")
        result = get_user_message_string(msg_info, None)
        expected = "Resource ID 'id' | Property 'prop' | msg"
        assert result == expected

    def test_with_field(self):
        msg_info = MessageInfo("msg", "id", field="field")
        result = get_user_message_string(msg_info, None)
        expected = "Resource ID 'id' | Field 'field' | msg"
        assert result == expected

    def test_without_res_id(self):
        msg_info = MessageInfo("msg", field="field")
        result = get_user_message_string(msg_info, None)
        expected = "Field 'field' | msg"
        assert result == expected

    def test_with_trace(self, message_info):
        result = get_user_message_string(message_info, "trace")
        expected = "File 'trace' | Resource ID 'id' | msg"
        assert result == expected

    def test_no_property(self, message_info):
        result = get_user_message_string(message_info, None)
        expected = "Resource ID 'id' | msg"
        assert result == expected


class TestStackFrame:
    def test_get_calling_code_context(self):
        result = _get_calling_code_context()
        assert isinstance(result, str)
        assert regex.search(r"test_xmllib_warnings_util\.py:\d+", result)

    def test_get_stack_frame_number(self):
        file_paths = [
            "/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/error/xmllib_warnings_util.py",
            "/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/error/xmllib_warnings_util.py",
            "/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/xmllib/models/dsp_base_resources.py",
            "/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/xmllib/models/dsp_base_resources.py",
            "<string>",
            "/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/xmllib/models/dsp_base_resources.py",
            "/Users/UserName/repoLocation/dsp-tools/test/unittests/xmllib/models/test_dsp_base_resources.py",
            "/Users/UserName/repoLocation/dsp-tools/.venv/lib/python3.12/site-packages/_pytest/python.py",
            "/Users/UserName/Applications/PyCharm CE.app/Contents/plugins/python-ce/helpers/pydev/pydevd.py",
        ]
        assert _get_stack_frame_number(file_paths) == 6

    def test_get_stack_frame_number_none(self):
        file_paths = [
            "/Users/UserName/repoLocation/dsp-tools/.venv/lib/python3.12/site-packages/_pytest/python.py",
            ("/Users/UserName/Applications/PyCharm CE.app/Contents/plugins/python-ce/helpers/pydev/pydevd.py"),
        ]
        assert _get_stack_frame_number(file_paths) == 0

    @pytest.mark.parametrize(
        ("f_path", "expected"),
        [
            ("/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/error/xmllib_warnings_util.py", True),
            ("/Users/UserName/repoLocation/dsp-tools/src/dsp_tools/xmllib/models/dsp_base_resources.py", True),
            ("<string>", True),
            ("/Users/UserName/repoLocation/dsp-tools/test/unittests/xmllib/models/test_dsp_base_resources.py", False),
            ("/Users/UserName/repoLocation/dsp-tools/.venv/lib/python3.12/site-packages/_pytest/python.py", False),
            (
                ("/Users/UserName/Applications/PyCharm CE.app/Contents/plugins/python-ce/helpers/pydev/pydevd.py"),
                False,
            ),
        ],
    )
    def test_filter_stack_frames(self, f_path, expected):
        assert _filter_stack_frames(f_path) == expected



from __future__ import annotations

from dataclasses import dataclass

from dsp_tools.commands.xmlupload.models.permission import Permissions
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedFileValue
from dsp_tools.commands.xmlupload.models.processed.file_values import ProcessedIIIFUri
from dsp_tools.commands.xmlupload.models.processed.values import ProcessedValue
from dsp_tools.legacy_models.datetimestamp import DateTimeStamp


@dataclass
class ProcessedResource:
    res_id: str
    type_iri: str
    label: str
    permissions: Permissions | None
    values: list[ProcessedValue]
    file_value: ProcessedFileValue | None = None
    iiif_uri: ProcessedIIIFUri | None = None
    migration_metadata: MigrationMetadata | None = None


@dataclass
class MigrationMetadata:
    iri_str: str | None
    creation_date: DateTimeStamp | None



from enum import Enum

from lxml import etree

from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.internal.constants import XML_NAMESPACE_MAP


class PermissionTypes(Enum):
    RV = "RV"
    V = "V"
    D = "D"
    CR = "CR"


class XMLPermissions:
    def serialise(self) -> list[etree._Element]:
        return [self._serialise_open(), self._serialise_restricted(), self._serialise_restricted_view()]

    def _serialise_open(self) -> etree._Element:
        permissions = {
            "UnknownUser": PermissionTypes.V,
            "KnownUser": PermissionTypes.V,
            "ProjectMember": PermissionTypes.D,
            "ProjectAdmin": PermissionTypes.CR,
        }
        return self._serialise_one_permission_element("open", permissions)

    def _serialise_restricted(self) -> etree._Element:
        permissions = {
            "ProjectMember": PermissionTypes.D,
            "ProjectAdmin": PermissionTypes.CR,
        }
        return self._serialise_one_permission_element("restricted", permissions)

    def _serialise_restricted_view(self) -> etree._Element:
        permissions = {
            "UnknownUser": PermissionTypes.RV,
            "KnownUser": PermissionTypes.RV,
            "ProjectMember": PermissionTypes.D,
            "ProjectAdmin": PermissionTypes.CR,
        }
        return self._serialise_one_permission_element("restricted-view", permissions)

    def _serialise_one_permission_element(
        self, permission_name: str, groups: dict[str, PermissionTypes]
    ) -> etree._Element:
        ele = etree.Element(f"{DASCH_SCHEMA}permissions", attrib={"id": permission_name}, nsmap=XML_NAMESPACE_MAP)
        allows = [self._serialise_one_allow(g, name) for g, name in groups.items()]
        ele.extend(allows)
        return ele

    def _serialise_one_allow(self, group: str, tag_text: PermissionTypes) -> etree._Element:
        ele = etree.Element(f"{DASCH_SCHEMA}allow", attrib={"group": group}, nsmap=XML_NAMESPACE_MAP)
        ele.text = tag_text.value
        return ele



# mypy: disable-error-code="method-assign,no-untyped-def"

import warnings
from pathlib import Path

import pandas as pd
import pytest
import regex

from dsp_tools.error.xmllib_warnings import XmllibInputWarning
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.file_values import FileValue
from dsp_tools.xmllib.models.internal.file_values import IIIFUri
from dsp_tools.xmllib.models.internal.file_values import Metadata
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended


@pytest.fixture
def metadata():
    return Metadata.new(
        LicenseRecommended.DSP.UNKNOWN,
        "copyright",
        ["authorship"],
        Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        "id",
    )


class TestMetadata:
    def test_good(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            Metadata.new(
                LicenseRecommended.DSP.UNKNOWN,
                "copyright",
                ["authorship"],
                Permissions.PROJECT_SPECIFIC_PERMISSIONS,
                "id",
            )
        assert len(caught_warnings) == 0

    def test_good_author_not_a_list(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            res = Metadata.new(
                LicenseRecommended.DSP.UNKNOWN,
                "copyright",
                "authorship",
                Permissions.PROJECT_SPECIFIC_PERMISSIONS,
                "id",
            )
        assert len(caught_warnings) == 0
        assert res.authorship == ("authorship",)

    def test_wrong_license(self):
        expected = regex.escape(
            "Field 'license (bistream/iiif-uri)' | "
            "The input should be a valid xmllib.License, your input 'license' does not match the type."
        )
        with pytest.warns(XmllibInputWarning, match=expected):
            Metadata.new("license", "copyright", ["authorship"], Permissions.PROJECT_SPECIFIC_PERMISSIONS, "id")  # type: ignore[arg-type]

    def test_wrong_copyright(self):
        expected = regex.escape(
            "Field 'copyright_holder (bistream/iiif-uri)' | Your input '' is empty. Please enter a valid string."
        )
        with pytest.warns(XmllibInputWarning, match=expected):
            Metadata.new(
                LicenseRecommended.DSP.UNKNOWN, "", ["authorship"], Permissions.PROJECT_SPECIFIC_PERMISSIONS, "id"
            )

    def test_wrong_authorship(self):
        expected = regex.escape(
            "Field 'authorship (bistream/iiif-uri)' | Your input '' is empty. Please enter a valid string."
        )
        with pytest.warns(XmllibInputWarning, match=expected):
            Metadata.new(
                LicenseRecommended.DSP.UNKNOWN, "copyright", [""], Permissions.PROJECT_SPECIFIC_PERMISSIONS, "id"
            )

    def test_wrong_permissions(self):
        expected = regex.escape(
            "Field 'permissions (bistream/iiif-uri)' | "
            "The input should be a valid xmllib.Permissions, your input 'string' does not match the type."
        )
        with pytest.warns(XmllibInputWarning, match=expected):
            Metadata.new(LicenseRecommended.DSP.UNKNOWN, "copyright", ["authorship"], "string", "id")  # type: ignore[arg-type]


class TestFileValue:
    def test_good(self, metadata):
        with warnings.catch_warnings(record=True) as caught_warnings:
            val = FileValue.new(
                Path("file.jpg"),
                metadata,
                pd.NA,  # type: ignore[arg-type]
                "id",
            )
        assert len(caught_warnings) == 0
        assert val.comment is None
        assert val.value == "file.jpg"

    def test_empty_path(self, metadata):
        expected = regex.escape("Field 'bitstream' | Your input '' is empty. Please enter a valid file name.")
        with pytest.warns(XmllibInputWarning, match=expected):
            val = FileValue.new(Path(""), metadata, None, "id")
        assert val.value == ""

    def test_path_is_none(self, metadata):
        expected = regex.escape("Field 'bitstream' | Your input 'None' is empty. Please enter a valid file name.")
        with pytest.warns(XmllibInputWarning, match=expected):
            val = FileValue.new(None, metadata, None, "id")  # type: ignore[arg-type]
        assert val.value == "None"


class TestIIIFUri:
    def test_good(self, metadata):
        with warnings.catch_warnings(record=True) as caught_warnings:
            val = IIIFUri.new(
                "https://example.org/image-service/abcd1234/full/max/0/default.jpg",
                metadata,
                pd.NA,  # type: ignore[arg-type]
                "id",
            )
        assert len(caught_warnings) == 0
        assert val.comment is None

    def test_wrong_val(self, metadata):
        expected = regex.escape("The input should be a valid IIIF uri, your input 'not a uri' does not match the type.")
        with pytest.warns(XmllibInputWarning, match=expected):
            IIIFUri.new("not a uri", metadata, None, "id")



import pytest

from dsp_tools.commands.excel2json.models.json_header import Descriptions
from dsp_tools.commands.excel2json.models.json_header import FilledJsonHeader
from dsp_tools.commands.excel2json.models.json_header import Keywords
from dsp_tools.commands.excel2json.models.json_header import Licenses
from dsp_tools.commands.excel2json.models.json_header import Prefixes
from dsp_tools.commands.excel2json.models.json_header import Project
from dsp_tools.commands.excel2json.models.json_header import User
from dsp_tools.commands.excel2json.models.json_header import UserRole
from dsp_tools.commands.excel2json.models.json_header import Users

SCHEMA = "https://raw.githubusercontent.com/dasch-swiss/dsp-tools/main/src/dsp_tools/resources/schema/project.json"


@pytest.fixture
def prefixes() -> Prefixes:
    return Prefixes({"dcelements": "http://purl.org/dc/elements/1.1/"})


@pytest.fixture
def descriptions() -> Descriptions:
    return Descriptions({"de": "Beschreibungstext", "en": "description text"})


@pytest.fixture
def keywords() -> Keywords:
    return Keywords(["Keyword 1"])


@pytest.fixture
def licenses() -> Licenses:
    return Licenses(["http://rdfh.ch/licenses/cc-by-4.0"])


@pytest.fixture
def user_sys_admin() -> User:
    return User("sys_admin", "sys_admin@email.ch", "given name1", "family name1", "PW1", "en", UserRole(sys_admin=True))


@pytest.fixture
def user_member() -> User:
    return User("member", "member@email.ch", "given name2", "family name2", "PW2", "de", UserRole())


@pytest.fixture
def user_admin() -> User:
    return User("admin", "admin@email.ch", "given name3", "family name3", "PW3", "de", UserRole(project_admin=True))


@pytest.fixture
def users(user_sys_admin: User, user_member: User, user_admin: User) -> Users:
    return Users([user_sys_admin, user_member, user_admin])


@pytest.fixture
def project_with_users(descriptions: Descriptions, keywords: Keywords, licenses: Licenses, users: Users) -> Project:
    return Project("0001", "shortname", "Longname of the project", descriptions, keywords, licenses, users)


@pytest.fixture
def project_no_users(descriptions: Descriptions, keywords: Keywords, licenses: Licenses) -> Project:
    return Project("0001", "shortname", "Longname of the project", descriptions, keywords, licenses, None)


@pytest.fixture
def project_no_licenses(descriptions: Descriptions, keywords: Keywords) -> Project:
    return Project("0001", "shortname", "Longname of the project", descriptions, keywords, Licenses([]), None)


@pytest.fixture
def filled_json_header_with_users_without_prefix(project_with_users: Project) -> FilledJsonHeader:
    return FilledJsonHeader(project_with_users, None)


@pytest.fixture
def filled_json_header_with_users_with_prefix(project_no_users: Project, prefixes: Prefixes) -> FilledJsonHeader:
    return FilledJsonHeader(project_no_users, prefixes)


def test_filled_json_header_with_users_without_prefix(
    filled_json_header_with_users_without_prefix: FilledJsonHeader,
) -> None:
    expected = {
        "$schema": SCHEMA,
        "project": {
            "shortcode": "0001",
            "shortname": "shortname",
            "longname": "Longname of the project",
            "descriptions": {"de": "Beschreibungstext", "en": "description text"},
            "keywords": ["Keyword 1"],
            "enabled_licenses": ["http://rdfh.ch/licenses/cc-by-4.0"],
            "users": [
                {
                    "username": "sys_admin",
                    "email": "sys_admin@email.ch",
                    "givenName": "given name1",
                    "familyName": "family name1",
                    "password": "PW1",
                    "lang": "en",
                    "status": True,
                    "groups": ["SystemAdmin"],
                    "projects": [":admin", ":member"],
                },
                {
                    "username": "member",
                    "email": "member@email.ch",
                    "givenName": "given name2",
                    "familyName": "family name2",
                    "password": "PW2",
                    "lang": "de",
                    "status": True,
                    "projects": [":member"],
                },
                {
                    "username": "admin",
                    "email": "admin@email.ch",
                    "givenName": "given name3",
                    "familyName": "family name3",
                    "password": "PW3",
                    "lang": "de",
                    "status": True,
                    "projects": [":admin", ":member"],
                },
            ],
        },
    }
    res = filled_json_header_with_users_without_prefix.to_dict()
    assert res == expected


def test_filled_json_header_with_users_with_prefix(filled_json_header_with_users_with_prefix: FilledJsonHeader) -> None:
    expected = {
        "prefixes": {"dcelements": "http://purl.org/dc/elements/1.1/"},
        "$schema": SCHEMA,
        "project": {
            "shortcode": "0001",
            "shortname": "shortname",
            "longname": "Longname of the project",
            "descriptions": {"de": "Beschreibungstext", "en": "description text"},
            "keywords": ["Keyword 1"],
            "enabled_licenses": ["http://rdfh.ch/licenses/cc-by-4.0"],
        },
    }
    res = filled_json_header_with_users_with_prefix.to_dict()
    assert res == expected


def test_filled_json_header_no_license(project_no_licenses: Project) -> None:
    header = FilledJsonHeader(project_no_licenses, None)
    expected = {
        "$schema": SCHEMA,
        "project": {
            "shortcode": "0001",
            "shortname": "shortname",
            "longname": "Longname of the project",
            "descriptions": {"de": "Beschreibungstext", "en": "description text"},
            "keywords": ["Keyword 1"],
            "enabled_licenses": [],
        },
    }
    res = header.to_dict()
    assert res == expected


if __name__ == "__main__":
    pytest.main([__file__])



from pathlib import Path

import regex
from lxml import etree

from dsp_tools.commands.validate_data.mappers import XML_TAG_TO_VALUE_TYPE_MAPPER
from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedMigrationMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue


def get_parsed_resources(root: etree._Element, api_url: str) -> list[ParsedResource]:
    api_url = _convert_api_url_for_correct_iri_namespace_construction(api_url)
    iri_lookup = _create_from_local_name_to_absolute_iri_lookup(root, api_url)
    all_res: list[ParsedResource] = []
    for res in root.iterdescendants(tag="resource"):
        res_type = iri_lookup[res.attrib["restype"]]
        all_res.append(_parse_one_resource(res, res_type, iri_lookup))
    for res in root.iterdescendants(tag="region"):
        res_type = f"{KNORA_API_STR}Region"
        all_res.append(_parse_one_resource(res, res_type, iri_lookup))
    for res in root.iterdescendants(tag="link"):
        res_type = f"{KNORA_API_STR}LinkObj"
        all_res.append(_parse_one_resource(res, res_type, iri_lookup))
    for res in root.iterdescendants(tag="video-segment"):
        all_res.append(_parse_segment(res, "Video"))
    for res in root.iterdescendants(tag="audio-segment"):
        all_res.append(_parse_segment(res, "Audio"))
    return all_res


def _convert_api_url_for_correct_iri_namespace_construction(api_url: str) -> str:
    return regex.sub(r"^https", "http", api_url)


def _create_from_local_name_to_absolute_iri_lookup(root: etree._Element, api_url: str) -> dict[str, str]:
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    local_names = {ele.attrib["restype"] for ele in root.iterdescendants(tag="resource")}
    props = {ele.attrib["name"] for ele in root.iter() if "name" in ele.attrib}
    local_names.update(props)
    lookup = {local: _get_one_absolute_iri(local, shortcode, default_ontology, api_url) for local in local_names}
    return lookup


def _get_one_absolute_iri(local_name: str, shortcode: str, default_ontology: str, api_url: str) -> str:
    split_name = local_name.split(":")
    if len(split_name) == 1:
        return f"{KNORA_API_STR}{local_name}"
    if len(split_name) == 2:
        if split_name[0] == "":
            return f"{_construct_namespace(api_url, shortcode, default_ontology)}{split_name[1]}"
        if split_name[0] == "knora-api":
            return f"{KNORA_API_STR}{split_name[1]}"
        return f"{_construct_namespace(api_url, shortcode, split_name[0])}{split_name[1]}"
    raise InputError(
        f"It is not permissible to have a colon in a property or resource class name. "
        f"Please correct the following: {local_name}"
    )


def _construct_namespace(api_url: str, shortcode: str, onto_name: str) -> str:
    return f"{api_url}/ontology/{shortcode}/{onto_name}/v2#"


def _parse_segment(segment: etree._Element, segment_type: str) -> ParsedResource:
    values = _parse_segment_values(segment, segment_type)
    migration_metadata = _parse_migration_metadata(segment)
    return ParsedResource(
        res_id=segment.attrib["id"],
        res_type=f"{KNORA_API_STR}{segment_type}Segment",
        label=segment.attrib["label"],
        permissions_id=segment.attrib.get("permissions"),
        values=values,
        file_value=None,
        migration_metadata=migration_metadata,
    )


def _parse_segment_values(segment: etree._Element, segment_type: str) -> list[ParsedValue]:
    values: list[ParsedValue] = []
    value: str | tuple[str, str] | None
    for val in segment.iterchildren():
        prop = f"{KNORA_API_STR}{val.tag!s}"
        match val.tag:
            case "isSegmentOf":
                val_type = KnoraValueType.LINK_VALUE
                prop = f"{KNORA_API_STR}is{segment_type}SegmentOf"
                value = val.text.strip() if val.text else None
            case "hasSegmentBounds":
                val_type = KnoraValueType.INTERVAL_VALUE
                value = (val.attrib["segment_start"], val.attrib["segment_end"])
            case "hasDescription" | "hasComment":
                val_type = KnoraValueType.RICHTEXT_VALUE
                value = _get_richtext_as_string(val)
            case "relatesTo":
                val_type = KnoraValueType.LINK_VALUE
                value = val.text.strip() if val.text else None
            case _:
                val_type = KnoraValueType.SIMPLETEXT_VALUE
                value = _get_simpletext_as_string(val)
        values.append(
            ParsedValue(
                prop_name=prop,
                value=value,
                value_type=val_type,
                permissions_id=val.attrib.get("permissions"),
                comment=val.attrib.get("comment"),
            )
        )
    return values


def _parse_one_resource(resource: etree._Element, res_type: str, iri_lookup: dict[str, str]) -> ParsedResource:
    values, file_value = _parse_values(resource, iri_lookup)
    migration_metadata = _parse_migration_metadata(resource)
    return ParsedResource(
        res_id=resource.attrib["id"],
        res_type=res_type,
        label=resource.attrib["label"],
        permissions_id=resource.attrib.get("permissions"),
        values=values,
        file_value=file_value,
        migration_metadata=migration_metadata,
    )


def _parse_migration_metadata(resource: etree._Element) -> ParsedMigrationMetadata | None:
    metadata = (resource.attrib.get("iri"), resource.attrib.get("ark"), resource.attrib.get("creation_date"))
    if any(metadata):
        return ParsedMigrationMetadata(
            iri=metadata[0],
            ark=metadata[1],
            creation_date=metadata[2],
        )
    return None


def _parse_values(
    resource: etree._Element, iri_lookup: dict[str, str]
) -> tuple[list[ParsedValue], ParsedFileValue | None]:
    values = []
    asset_value = None
    for val in resource.iterchildren():
        match val.tag:
            case "bitstream":
                asset_value = _parse_file_values(val)
            case "iiif-uri":
                asset_value = _parse_iiif_uri(val)
            case _:
                values.extend(_parse_one_value(val, iri_lookup))
    return values, asset_value


def _parse_one_value(values: etree._Element, iri_lookup: dict[str, str]) -> list[ParsedValue]:
    prop_name = iri_lookup[values.attrib["name"]]
    match values.tag:
        case "list-prop":
            return _parse_list_value(values, prop_name)
        case "text-prop":
            return _parse_text_value(values, prop_name)
        case _:
            return _parse_generic_values(values, prop_name)


def _parse_generic_values(values: etree._Element, prop_name: str) -> list[ParsedValue]:
    value_type = XML_TAG_TO_VALUE_TYPE_MAPPER[str(values.tag)]
    parsed_values = []
    for val in values:
        parsed_values.append(
            ParsedValue(
                prop_name=prop_name,
                value=val.text.strip() if val.text else None,
                value_type=value_type,
                permissions_id=val.attrib.get("permissions"),
                comment=val.attrib.get("comment"),
            )
        )
    return parsed_values


def _parse_list_value(values: etree._Element, prop_name: str) -> list[ParsedValue]:
    parsed_values = []
    list_name = values.attrib["list"]
    for val in values:
        list_node = val.text.strip() if val.text else None
        parsed_values.append(
            ParsedValue(
                prop_name=prop_name,
                value=(list_name, list_node),
                value_type=KnoraValueType.LIST_VALUE,
                permissions_id=val.attrib.get("permissions"),
                comment=val.attrib.get("comment"),
            )
        )
    return parsed_values


def _parse_text_value(values: etree._Element, prop_name: str) -> list[ParsedValue]:
    parsed_values = []
    for val in values:
        if val.attrib["encoding"] == "xml":
            val_type = KnoraValueType.RICHTEXT_VALUE
            value = _get_richtext_as_string(val)
        else:
            val_type = KnoraValueType.SIMPLETEXT_VALUE
            value = _get_simpletext_as_string(val)
        parsed_values.append(
            ParsedValue(
                prop_name=prop_name,
                value=value,
                value_type=val_type,
                permissions_id=val.attrib.get("permissions"),
                comment=val.attrib.get("comment"),
            )
        )
    return parsed_values


def _get_richtext_as_string(value: etree._Element) -> str | None:
    # Not entering any values within the tag results in None,
    # however if only whitespaces are entered then it should return an empty string so that the user message is precise.
    if not value.text and len(value) == 0:
        return None
    xmlstr = etree.tostring(value, encoding="unicode", method="xml").strip()
    xmlstr = regex.sub(f"^<{value.tag!s}.*?>", "", xmlstr, count=1)
    xmlstr = regex.sub(f"</{value.tag!s}>$", "", xmlstr)
    striped_str = xmlstr.strip()
    return _cleanup_formatted_text(striped_str)


def _cleanup_formatted_text(xmlstr_orig: str) -> str:
    """
    In a xml-encoded text value from the XML file,
    there may be non-text characters that must be removed.
    This function:
        - replaces (multiple) line breaks by a space
        - replaces multiple spaces or tabstops by a single space (except within `<code>` or `<pre>` tags)

    Args:
        xmlstr_orig: content of the tag from the XML file, in serialized form

    Returns:
        purged string, suitable to be sent to DSP-API
    """
    # replace (multiple) line breaks by a space
    xmlstr = regex.sub("\n+", " ", xmlstr_orig)
    # replace multiple spaces or tabstops by a single space (except within <code> or <pre> tags)
    # the regex selects all spaces/tabstops not followed by </xyz> without <xyz in between.
    # credits: https://stackoverflow.com/a/46937770/14414188
    xmlstr = regex.sub("( {2,}|\t+)(?!(.(?!<(code|pre)))*</(code|pre)>)", " ", xmlstr)
    # remove spaces after <br/> tags (except within <code> tags)
    xmlstr = regex.sub("((?<=<br/?>) )(?!(.(?!<code))*</code>)", "", xmlstr)
    # remove leading and trailing spaces
    xmlstr = xmlstr.strip()
    return xmlstr


def _get_simpletext_as_string(value: etree._Element) -> str | None:
    # Not entering any values within the tag results in None,
    # however if only whitespaces are entered then it should return an empty string so that the user message is precise.
    if len(value) == 0:
        if not (found := value.text):
            return None
    else:
        # Extract the inner XML content, preserving tags
        found = "".join(etree.tostring(child, encoding="unicode") for child in value.iterdescendants())
        if value.text:
            found = value.text + found
    # replace multiple spaces or tabstops by a single space
    str_val = regex.sub(r" {2,}|\t+", " ", found)
    # remove leading and trailing spaces (of every line, but also of the entire string)
    str_val = "\n".join([s.strip() for s in str_val.split("\n")])
    result = str_val.strip()
    return result


def _parse_iiif_uri(iiif_uri: etree._Element) -> ParsedFileValue:
    return ParsedFileValue(
        value=iiif_uri.text.strip() if iiif_uri.text else None,
        value_type=KnoraValueType.STILL_IMAGE_IIIF,
        metadata=_parse_file_metadata(iiif_uri),
    )


def _parse_file_values(file_value: etree._Element) -> ParsedFileValue:
    val = file_value.text.strip() if file_value.text else None
    return ParsedFileValue(
        value=val,
        value_type=_get_file_value_type(val),
        metadata=_parse_file_metadata(file_value),
    )


def _parse_file_metadata(file_value: etree._Element) -> ParsedFileValueMetadata:
    return ParsedFileValueMetadata(
        license_iri=file_value.attrib.get("license"),
        copyright_holder=file_value.attrib.get("copyright-holder"),
        authorship_id=file_value.attrib.get("authorship-id"),
        permissions_id=file_value.attrib.get("permissions"),
    )


def _get_file_value_type(file_name: str | None) -> KnoraValueType | None:  # noqa:PLR0911 (Too many return statements)
    if not file_name:
        return None
    file_extension = Path(file_name).suffix[1:].lower()
    match file_extension:
        case "zip" | "tar" | "gz" | "z" | "tgz" | "gzip" | "7z":
            return KnoraValueType.ARCHIVE_FILE
        case "mp3" | "wav":
            return KnoraValueType.AUDIO_FILE
        case "pdf" | "doc" | "docx" | "xls" | "xlsx" | "ppt" | "pptx":
            return KnoraValueType.DOCUMENT_FILE
        case "mp4":
            return KnoraValueType.MOVING_IMAGE_FILE
        # jpx is the extension of the files returned by dsp-ingest
        case "jpg" | "jpeg" | "jp2" | "png" | "tif" | "tiff" | "jpx":
            return KnoraValueType.STILL_IMAGE_FILE
        case "odd" | "rng" | "txt" | "xml" | "xsd" | "xsl" | "csv" | "json":
            return KnoraValueType.TEXT_FILE
        case _:
            return None



from __future__ import annotations

import datetime
from argparse import ArgumentParser
from argparse import _SubParsersAction
from importlib.metadata import version

# help texts
username_text = "username (e-mail) used for authentication with the DSP-API"
password_text = "password used for authentication with the DSP-API"
dsp_server_text = "URL of the DSP server"
verbose_text = "print more information about the progress to the console"


def make_parser(
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> ArgumentParser:
    """
    Create a parser for the command line arguments.

    Args:
        default_dsp_api_url: URL of the DSP server (default value for localhost)
        root_user_email: username (e-mail) used for authentication with the DSP-API (default value for localhost)
        root_user_pw: password used for authentication with the DSP-API (default value for localhost)

    Returns:
        parser
    """

    parser = ArgumentParser(
        description=f"DSP-TOOLS (version {version('dsp-tools')}, © {datetime.datetime.now().year} by DaSCH)"
    )
    subparsers = parser.add_subparsers(
        title="Subcommands", description="Valid subcommands are", help="sub-command help"
    )

    _add_start_stack(subparsers)

    _add_stop_stack(subparsers)

    _add_create(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_get(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_xmlupload(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_validate_data(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_resume_xmlupload(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_upload_files(subparsers, default_dsp_api_url, root_user_email, root_user_pw)
    _add_ingest_files(subparsers, default_dsp_api_url, root_user_email, root_user_pw)
    _add_ingest_xmlupload(subparsers, default_dsp_api_url, root_user_email, root_user_pw)

    _add_excel2json(subparsers)
    _add_old_excel2json(subparsers)

    _add_excel2lists(subparsers)
    _add_old_excel2lists(subparsers)

    _add_excel2resources(subparsers)

    _add_excel2properties(subparsers)

    _add_excel2xml(subparsers)

    _add_id2iri(subparsers)

    _add_create_template(subparsers)

    _add_rosetta(subparsers)

    _add_suppress_update_prompt(subparsers)

    return parser


def _add_suppress_update_prompt(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    outdated_help_text = (
        "don't prompt when using an outdated version of DSP-TOOLS "
        "(useful for contexts without interactive shell, e.g. when the Terminal output is piped into a file)"
    )
    for sp in subparsers.choices.values():
        sp.add_argument("--suppress-update-prompt", action="store_true", help=outdated_help_text)


def _add_rosetta(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="rosetta", help="Clone the most up to data rosetta repository, create the data model and upload the data"
    )
    subparser.set_defaults(action="rosetta")


def _add_create_template(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    # create template repo with minimal JSON and XML files
    subparser = subparsers.add_parser(
        name="template", help="Create a template repository with a minimal JSON and XML file"
    )
    subparser.set_defaults(action="template")


def _add_stop_stack(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="stop-stack", help="Shut down the local instance of DSP-API and DSP-APP, and delete all data in it"
    )
    subparser.set_defaults(action="stop-stack")


def _add_start_stack(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    latest_text = (
        "instead of the latest deployed version, "
        "use the latest development version (from the `main` branch) "
        "of the backend components (api, sipi, fuseki, ingest)"
    )
    max_file_size_text = "max. multimedia file size allowed for ingest, in MB (default: 2000, max: 100'000)"
    no_prune_text = "don't execute 'docker system prune' (and don't ask)"
    with_test_data_text = "initialise the database with built-in test data"
    custom_host = "set host to use stack on a server"
    subparser = subparsers.add_parser(name="start-stack", help="Run a local instance of DSP-API and DSP-APP")
    subparser.set_defaults(action="start-stack")
    subparser.add_argument("--max_file_size", type=int, help=max_file_size_text)
    subparser.add_argument("--prune", action="store_true", help="execute 'docker system prune' without asking")
    subparser.add_argument("--no-prune", action="store_true", help=no_prune_text)
    subparser.add_argument("--latest", action="store_true", help=latest_text)
    subparser.add_argument("--with-test-data", action="store_true", help=with_test_data_text)
    subparser.add_argument("--custom-host", default=None, type=str, help=custom_host)


def _add_id2iri(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="id2iri",
        help="Replace internal IDs of an XML file (resptr tags or salsah-links) by IRIs provided in a mapping file.",
    )
    subparser.set_defaults(action="id2iri")
    subparser.add_argument(
        "-r", "--remove-resources", action="store_true", help="remove resources if their ID is in the mapping"
    )
    subparser.add_argument("xmlfile", help="path to the XML file containing the data to be replaced")
    subparser.add_argument("mapping", help="path to the JSON file containing the mapping of IDs to IRIs")


def _add_excel2xml(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="excel2xml",
        help="Create an XML file from an Excel/CSV file that is already structured according to the DSP specifications",
    )
    subparser.set_defaults(action="excel2xml")
    subparser.add_argument("data_source", help="path to the CSV or XLS(X) file containing the data")
    subparser.add_argument("project_shortcode", help="shortcode of the project that this data belongs to")
    subparser.add_argument("ontology_name", help="name of the ontology that this data belongs to")


def _add_excel2properties(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="excel2properties",
        help="Create the 'properties' section of a JSON project file from one or multiple Excel files",
    )
    subparser.set_defaults(action="excel2properties")
    subparser.add_argument("excelfile", help="path to the Excel file containing the properties")
    subparser.add_argument(
        "properties_section", help="path to the output JSON file containing the 'properties' section"
    )


def _add_excel2resources(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="excel2resources",
        help="Create the 'resources' section of a JSON project file from one or multiple Excel files",
    )
    subparser.set_defaults(action="excel2resources")
    subparser.add_argument("excelfile", help="path to the Excel file containing the resources")
    subparser.add_argument("resources_section", help="path to the output JSON file containing the 'resources' section")


def _add_excel2lists(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="excel2lists",
        help="Create the 'lists' section of a JSON project file from one or multiple Excel files",
    )
    subparser.set_defaults(action="excel2lists")
    subparser.add_argument("excelfolder", help="path to the folder containing the Excel file(s)")
    subparser.add_argument("lists_section", help="path to the output JSON file containing the 'lists' section")


def _add_old_excel2lists(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="old-excel2lists",
        help="DEPRECATED: Create the 'lists' section of a JSON project file from one or multiple Excel files",
    )
    subparser.set_defaults(action="old-excel2lists")
    subparser.add_argument("-v", "--verbose", action="store_true", help=verbose_text)
    subparser.add_argument("excelfolder", help="path to the folder containing the Excel file(s)")
    subparser.add_argument("lists_section", help="path to the output JSON file containing the 'lists' section")


def _add_old_excel2json(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="old-excel2json",
        help="DEPRECATED: Create an entire JSON project file from a folder containing the required Excel files",
    )
    subparser.set_defaults(action="old-excel2json")
    subparser.add_argument("excelfolder", help="path to the folder containing the Excel files")
    subparser.add_argument("project_definition", help="path to the output JSON file")


def _add_excel2json(subparsers: _SubParsersAction[ArgumentParser]) -> None:
    subparser = subparsers.add_parser(
        name="excel2json",
        help="Create an entire JSON project file from a folder containing the required Excel files",
    )
    subparser.set_defaults(action="excel2json")
    subparser.add_argument("excelfolder", help="path to the folder containing the Excel files")
    subparser.add_argument("project_definition", help="path to the output JSON file")


def _add_upload_files(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(
        name="upload-files",
        help="Experimental: upload all files referenced in an XML file to a DSP server",
    )
    subparser.set_defaults(action="upload-files")
    subparser.add_argument("-s", "--server", default=default_dsp_api_url, help=dsp_server_text)
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument(
        "-i", "--imgdir", default=".", help="folder from where the paths in the <bitstream> tags are evaluated"
    )
    subparser.add_argument("xml_file", help="path to XML file containing the data")


def _add_ingest_files(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(
        name="ingest-files",
        help="Experimental: ingest uploaded files on a DSP server and retrieve the mapping CSV file",
    )
    subparser.set_defaults(action="ingest-files")
    subparser.add_argument("-s", "--server", default=default_dsp_api_url, help=dsp_server_text)
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument("shortcode", help="shortcode of the project that the files belong to")


def _add_ingest_xmlupload(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(
        name="ingest-xmlupload",
        help="For internal use only: create resources with files already uploaded through dsp-ingest",
    )
    subparser.set_defaults(action="ingest-xmlupload")
    subparser.add_argument("-s", "--server", default=default_dsp_api_url, help=dsp_server_text)
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument("--interrupt-after", type=int, default=-1, help="interrupt after this number of resources")
    subparser.add_argument("xml_file", help="path to XML file containing the data")


def _add_xmlupload(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(name="xmlupload", help="Upload data defined in an XML file to a DSP server")
    subparser.set_defaults(action="xmlupload")
    subparser.add_argument(
        "-s", "--server", default=default_dsp_api_url, help="URL of the DSP server where DSP-TOOLS sends the data to"
    )
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument(
        "-i", "--imgdir", default=".", help="folder from where the paths in the <bitstream> tags are evaluated"
    )
    subparser.add_argument(
        "-V", "--validate-only", action="store_true", help="validate the XML file without uploading it"
    )
    subparser.add_argument("--interrupt-after", type=int, default=-1, help="interrupt after this number of resources")
    subparser.add_argument("xmlfile", help="path to the XML file containing the data")
    subparser.add_argument("--no-iiif-uri-validation", action="store_true", help="skip the IIIF URI validation")
    subparser.add_argument(
        "--validation-severity",
        choices=["error", "warning", "info"],
        help="Which severity level of validation message should be printed out",
        default="info",
    )


def _add_validate_data(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(name="validate-data", help="Validate the data with the data model.")
    subparser.set_defaults(action="validate-data")
    subparser.add_argument("xmlfile", help="path to the XML file containing the data")
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument(
        "-s", "--server", default=default_dsp_api_url, help="URL of the DSP server where DSP-TOOLS sends the data to"
    )

    subparser.add_argument(
        "--save-graphs", action="store_true", help="Save the data, onto and shacl graph as ttl files."
    )


def _add_resume_xmlupload(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(name="resume-xmlupload", help="Resume an interrupted xmlupload")
    subparser.set_defaults(action="resume-xmlupload")
    subparser.add_argument(
        "-s", "--server", default=default_dsp_api_url, help="URL of the DSP server where DSP-TOOLS sends the data to"
    )
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    skip_text = (
        "In case the last resource was successfully uploaded to the DSP-API, "
        "but DSP-TOOLS did not get a confirmation from the DSP-API, "
        "the resource will remain in the pending upload list.\n"
        "If this flag is set, the first resource in the list is skipped."
    )
    subparser.add_argument("--skip-first-resource", action="store_true", help=skip_text)


def _add_get(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(
        name="get",
        help="Retrieve a project with its data model(s) from a DSP server and write it into a JSON file",
    )
    subparser.set_defaults(action="get")
    subparser.add_argument("-s", "--server", default=default_dsp_api_url, help=dsp_server_text)
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument("-P", "--project", help="shortcode, shortname or IRI of the project", required=True)
    subparser.add_argument("-v", "--verbose", action="store_true", help=verbose_text)
    subparser.add_argument("project_definition", help="path to the file the project should be written to")


def _add_create(
    subparsers: _SubParsersAction[ArgumentParser],
    default_dsp_api_url: str,
    root_user_email: str,
    root_user_pw: str,
) -> None:
    subparser = subparsers.add_parser(
        name="create",
        help="Create a project defined in a JSON project file on a DSP server. "
        "A project can consist of lists, groups, users, and ontologies (data models).",
    )
    subparser.set_defaults(action="create")
    subparser.add_argument("-s", "--server", default=default_dsp_api_url, help=dsp_server_text)
    subparser.add_argument("-u", "--user", default=root_user_email, help=username_text)
    subparser.add_argument("-p", "--password", default=root_user_pw, help=password_text)
    subparser.add_argument(
        "-V",
        "--validate-only",
        action="store_true",
        help="validate the JSON file without creating it on the DSP server",
    )
    subparser.add_argument(
        "-l",
        "--lists-only",
        action="store_true",
        help="create only the lists (prerequisite: the project exists on the server)",
    )
    subparser.add_argument("-v", "--verbose", action="store_true", help=verbose_text)
    subparser.add_argument("project_definition", help="path to the JSON project file")



from __future__ import annotations

import warnings
from pathlib import Path

import pandas as pd
import regex

from dsp_tools.commands.excel2json.models.input_error import AtLeastOneValueRequiredProblem
from dsp_tools.commands.excel2json.models.input_error import EmptySheetsProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelSheetProblem
from dsp_tools.commands.excel2json.models.input_error import InvalidExcelContentProblem
from dsp_tools.commands.excel2json.models.input_error import MissingValuesProblem
from dsp_tools.commands.excel2json.models.input_error import MoreThanOneRowProblem
from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import RequiredColumnMissingProblem
from dsp_tools.commands.excel2json.models.json_header import Descriptions
from dsp_tools.commands.excel2json.models.json_header import EmptyJsonHeader
from dsp_tools.commands.excel2json.models.json_header import FilledJsonHeader
from dsp_tools.commands.excel2json.models.json_header import JsonHeader
from dsp_tools.commands.excel2json.models.json_header import Keywords
from dsp_tools.commands.excel2json.models.json_header import Licenses
from dsp_tools.commands.excel2json.models.json_header import Prefixes
from dsp_tools.commands.excel2json.models.json_header import Project
from dsp_tools.commands.excel2json.models.json_header import User
from dsp_tools.commands.excel2json.models.json_header import UserRole
from dsp_tools.commands.excel2json.models.json_header import Users
from dsp_tools.commands.excel2json.utils import check_contains_required_columns
from dsp_tools.commands.excel2json.utils import find_missing_required_values
from dsp_tools.commands.excel2json.utils import read_and_clean_all_sheets
from dsp_tools.error.custom_warnings import DspToolsFutureWarning
from dsp_tools.error.exceptions import InputError
from dsp_tools.error.problems import Problem
from dsp_tools.utils.data_formats.uri_util import is_uri


def get_json_header(excel_filepath: Path) -> JsonHeader:
    """
    Returns the header of the JSON file.
    If the Excel file exists, the header will be filled in.
    Otherwise, it will return a header with the fields left blank.

    Args:
        excel_filepath: path to the excel file

    Returns:
        JsonHeader object

    Raises:
        InputError: If the file does not conform to the specifications
    """
    if not excel_filepath.is_file():
        print("No json_header.xlsx file found in the directory, empty header was added.")
        return EmptyJsonHeader()
    sheets_df_dict = read_and_clean_all_sheets(excel_filepath)
    sheets_df_dict = {k.lower(): df for k, df in sheets_df_dict.items()}
    if compliance_problem := _do_all_checks(sheets_df_dict):
        raise InputError(compliance_problem.execute_error_protocol())
    print("json_header.xlsx file used to construct the json header.")
    return _process_file(sheets_df_dict)


def _do_all_checks(df_dict: dict[str, pd.DataFrame]) -> ExcelFileProblem | None:
    _futurewarning_about_inexistent_license_info(df_dict)
    if file_problems := _check_if_sheets_are_filled_and_exist(df_dict):
        return file_problems
    sheet_problems: list[Problem] = []
    if prefix_problem := _check_prefixes(df_dict["prefixes"]):
        sheet_problems.append(prefix_problem)
    if project_problems := _check_project_sheet(df_dict["project"]):
        sheet_problems.append(project_problems)
    if description_problem := _check_descriptions(df_dict["description"]):
        sheet_problems.append(description_problem)
    if keywords_problem := _check_keywords(df_dict["keywords"]):
        sheet_problems.append(keywords_problem)
    if (licenses_df := df_dict.get("licenses")) is not None:
        if license_problem := _check_licenses(licenses_df):
            sheet_problems.append(license_problem)
    if (user_df := df_dict.get("users")) is not None:
        if user_problems := _check_all_users(user_df):
            sheet_problems.append(user_problems)
    if sheet_problems:
        return ExcelFileProblem("json_header.xlsx", sheet_problems)
    return None


def _futurewarning_about_inexistent_license_info(df_dict: dict[str, pd.DataFrame]) -> None:
    if df_dict.get("licenses") is None:
        msg = (
            "The json_header.xlsx file does not have a sheet containing the enabled licenses. "
            "Please note that this will become mandatory in the future."
        )
        warnings.warn(DspToolsFutureWarning(msg))


def _check_if_sheets_are_filled_and_exist(df_dict: dict[str, pd.DataFrame]) -> ExcelFileProblem | None:
    expected_sheets = ["prefixes", "project", "description", "keywords"]

    def _check_df(sheet: str) -> str | None:
        if (df := df_dict.get(sheet)) is None:
            return sheet
        if len(df) == 0:
            return sheet
        return None

    if empty_or_missing_sheets := [x for x in expected_sheets if _check_df(x)]:
        return ExcelFileProblem("json_header.xlsx", [EmptySheetsProblem(empty_or_missing_sheets)])
    return None


def _check_prefixes(df: pd.DataFrame) -> ExcelSheetProblem | None:
    if missing_cols := check_contains_required_columns(df, {"prefixes", "uri"}):
        return ExcelSheetProblem("prefixes", [missing_cols])
    problems: list[Problem] = []
    if missing_vals := find_missing_required_values(df, ["prefixes", "uri"]):
        problems.append(MissingValuesProblem(missing_vals))
    if not (uri_series := pd.Series([is_uri(x) for x in df["uri"].tolist()])).all():
        problematic_uri = df["uri"][~uri_series]
        problematic_vals: list[Problem] = [
            InvalidExcelContentProblem(
                expected_content="A valid URI",
                actual_content=value,
                excel_position=PositionInExcel(column="uri", row=int(str(i)) + 2),
            )
            for i, value in problematic_uri.items()
        ]
        problems.extend(problematic_vals)
    if problems:
        return ExcelSheetProblem("prefixes", problems)
    return None


def _check_project_sheet(df: pd.DataFrame) -> ExcelSheetProblem | None:
    problems: list[Problem] = []
    cols = {"shortcode", "shortname", "longname"}
    if missing_cols := check_contains_required_columns(df, cols):
        problems.append(missing_cols)
    if len(df) > 1:
        problems.append(MoreThanOneRowProblem(len(df)))
    if problems:
        return ExcelSheetProblem("project", problems)
    if missing_values := find_missing_required_values(df, list(cols)):
        return ExcelSheetProblem("project", [MissingValuesProblem(missing_values)])
    return None


def _check_descriptions(df: pd.DataFrame) -> ExcelSheetProblem | None:
    if len(df) > 1:
        return ExcelSheetProblem("description", [MoreThanOneRowProblem(len(df))])
    desc_columns = ["description_en", "description_de", "description_fr", "description_it", "description_rm"]
    extracted_desc = _extract_descriptions(df)
    if not extracted_desc.descriptions:
        return ExcelSheetProblem("description", [AtLeastOneValueRequiredProblem(desc_columns, 2)])
    return None


def _check_keywords(df: pd.DataFrame) -> ExcelSheetProblem | None:
    if "keywords" not in df.columns:
        return ExcelSheetProblem("keywords", [RequiredColumnMissingProblem(["keywords"])])
    extracted_keywords = _extract_keywords(df)
    if len(extracted_keywords.keywords) == 0:
        return ExcelSheetProblem("keywords", [MissingValuesProblem([PositionInExcel(column="keywords")])])
    return None


def _check_licenses(df: pd.DataFrame) -> ExcelSheetProblem | None:
    if missing_cols := check_contains_required_columns(df, {"enabled"}):
        return ExcelSheetProblem("licenses", [missing_cols])
    return None


def _check_all_users(df: pd.DataFrame) -> ExcelSheetProblem | None:
    if not len(df) > 0:
        return None
    columns = ["username", "email", "givenname", "familyname", "password", "lang", "role"]
    if missing_cols := check_contains_required_columns(df, set(columns)):
        return ExcelSheetProblem("users", [missing_cols])
    if missing_vals := find_missing_required_values(df, columns):
        return ExcelSheetProblem("users", [MissingValuesProblem(missing_vals)])
    problems: list[Problem] = []
    for i, row in df.iterrows():
        if user_problem := _check_one_user(row, int(str(i)) + 2):
            problems.extend(user_problem)
    if problems:
        return ExcelSheetProblem("users", problems)
    return None


def _check_one_user(row: pd.Series[str], row_number: int) -> list[InvalidExcelContentProblem]:
    problems: list[InvalidExcelContentProblem] = []
    if bad_language := _check_lang(row["lang"], row_number):
        problems.append(bad_language)
    if bad_email := _check_email(row["email"], row_number):
        problems.append(bad_email)
    if bad_role := _check_role(row["role"], row_number):
        problems.append(bad_role)
    return problems


def _check_lang(value: str, row_num: int) -> InvalidExcelContentProblem | None:
    lang_options = ["en", "de", "fr", "it", "rm"]
    if value.lower() in lang_options:
        return None
    return InvalidExcelContentProblem(
        expected_content="One of: en, de, fr, it, rm",
        actual_content=value,
        excel_position=PositionInExcel(column="lang", row=row_num),
    )


def _check_email(email: str, row_num: int) -> InvalidExcelContentProblem | None:
    if not regex.search(r".+@.+\..+", email):
        return InvalidExcelContentProblem(
            expected_content="A valid email adress",
            actual_content=email,
            excel_position=PositionInExcel(column="email", row=row_num),
        )
    return None


def _check_role(value: str, row_num: int) -> InvalidExcelContentProblem | None:
    possible_roles = ["projectadmin", "systemadmin", "projectmember"]
    if value.lower() not in possible_roles:
        return InvalidExcelContentProblem(
            expected_content="One of: projectadmin, systemadmin, projectmember",
            actual_content=value,
            excel_position=PositionInExcel(column="role", row=row_num),
        )
    return None


def _process_file(df_dict: dict[str, pd.DataFrame]) -> FilledJsonHeader:
    prefixes = _extract_prefixes(df_dict["prefixes"])
    project = _extract_project(df_dict)
    return FilledJsonHeader(project, prefixes)


def _extract_prefixes(df: pd.DataFrame) -> Prefixes:
    pref: dict[str, str] = dict(zip(df["prefixes"], df["uri"]))
    pref = {k.rstrip(":"): v for k, v in pref.items()}
    return Prefixes(pref)


def _extract_project(df_dict: dict[str, pd.DataFrame]) -> Project:
    project_df = df_dict["project"]
    extracted_description = _extract_descriptions(df_dict["description"])
    extracted_keywords = _extract_keywords(df_dict["keywords"])
    if (lic_df := df_dict.get("licenses")) is not None:
        extracted_licenses = _extract_licenses(lic_df)
    else:
        extracted_licenses = Licenses([])
    all_users = None
    if (user_df := df_dict.get("users")) is not None:
        if len(user_df) > 0:
            all_users = _extract_users(user_df)
    shortcode = str(project_df.loc[0, "shortcode"]).zfill(4)
    return Project(
        shortcode=shortcode,
        shortname=str(project_df.loc[0, "shortname"]),
        longname=str(project_df.loc[0, "longname"]),
        descriptions=extracted_description,
        keywords=extracted_keywords,
        licenses=extracted_licenses,
        users=all_users,
    )


def _extract_descriptions(df: pd.DataFrame) -> Descriptions:
    desc_col_dict = _get_description_cols(list(df.columns))
    return Descriptions(
        {lang: str(value) for lang, column in desc_col_dict.items() if not pd.isna(value := df.loc[0, column])}
    )


def _get_description_cols(cols: list[str]) -> dict[str, str]:
    re_pat = r"^(\w*)(en|it|de|fr|rm)$"
    return {found.group(2): x for x in cols if (found := regex.search(re_pat, x))}


def _extract_keywords(df: pd.DataFrame) -> Keywords:
    keywords = list({x for x in df["keywords"] if not pd.isna(x)})
    return Keywords(keywords)


def _extract_licenses(df: pd.DataFrame) -> Licenses:
    licenses = list({x for x in df["enabled"] if not pd.isna(x)})
    return Licenses(licenses)


def _extract_users(df: pd.DataFrame) -> Users:
    users = []
    for _, row in df.iterrows():
        str_row: pd.Series[str] = row
        users.append(_extract_one_user(str_row))
    return Users(users)


def _extract_one_user(row: pd.Series[str]) -> User:
    user_role = _get_role(row["role"])
    return User(
        username=row["username"],
        email=row["email"],
        givenName=row["givenname"],
        familyName=row["familyname"],
        password=row["password"],
        lang=row["lang"],
        role=user_role,
    )


def _get_role(value: str) -> UserRole:
    match value.lower():
        case "projectadmin":
            return UserRole(project_admin=True)
        case "systemadmin":
            return UserRole(sys_admin=True)
        case _:
            return UserRole()



from __future__ import annotations

from dataclasses import dataclass

from dsp_tools.commands.xmlupload.models.permission import Permissions


@dataclass(frozen=True)
class BitstreamInfo:
    """
    Represents a bitstream object,
    consisting of its file name on the local file system,
    the internal file name assigned by the ingest service
    and optionally its permissions.
    """

    internal_file_name: str
    permissions: Permissions | None = None



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from enum import auto
from uuid import uuid4

from dsp_tools.legacy_models.datetimestamp import DateTimeStamp
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType


@dataclass
class RdfLikeData:
    resources: list[RdfLikeResource]


@dataclass
class RdfLikeResource:
    """
    Represents a user facing project-specific resource.

    Args:
        res_id: resource ID provided by the user in the XML
        property_objects: A list of properties and objects where the subject is the resource itself.
            They are non-reified triples (not values).
            For example, the label of a resource is here.
        values: a list of user-facing values (e.g. BooleanValue)
        asset_value: an asset value (all FileValues excluding the IIIFUri)
        migration_metadata: Metadata used for salsah migration
    """

    res_id: str
    property_objects: list[PropertyObject]
    values: list[RdfLikeValue]
    asset_value: RdfLikeValue | None
    migration_metadata: MigrationMetadata


@dataclass
class MigrationMetadata:
    iri: str | None = None
    ark: str | None = None
    creation_date: DateTimeStamp | None = None

    def any(self) -> bool:
        return any([self.iri, self.ark, self.creation_date])


@dataclass
class PropertyObject:
    """
    Property and object of a triple.

    Args:
        property_type: maps to a specific knora-api or rdf(s) property
        object_value: object of the triple, may be user facing (e.g. label) or metadata (e.g. permissions)
        object_type: datatype for literals (e.g. boolean) or that it is an IRI (not a literal)
    """

    property_type: TriplePropertyType
    object_value: str | None
    object_type: TripleObjectType


@dataclass
class RdfLikeValue:
    """
    Contains information about a user-facing value, for example BooleanValue.

    Args:
        user_facing_prop: Absolute IRI of the property as defined in the ontology
        user_facing_value: User-facing value, for example a number
        knora_type: Maps to a knora value type (e.g. BooleanValue)
        value_metadata: metadata of the value for example permissions, comments, etc.
    """

    user_facing_prop: str
    user_facing_value: str | None
    knora_type: KnoraValueType
    value_metadata: list[PropertyObject]
    value_uuid: str = field(default_factory=lambda: str(uuid4()))


class TriplePropertyType(Enum):
    """
    Maps to a specific knora-api or rdf(s) property.
    For example: KNORA_COMMENT_ON_VALUE -> knora-api:valueHasComment
    """

    RDFS_LABEL = auto()
    RDF_TYPE = auto()
    KNORA_PERMISSIONS = auto()
    KNORA_COMMENT_ON_VALUE = auto()
    KNORA_INTERVAL_START = auto()
    KNORA_INTERVAL_END = auto()
    KNORA_STANDOFF_LINK = auto()
    KNORA_LICENSE = auto()
    KNORA_AUTHORSHIP = auto()
    KNORA_COPYRIGHT_HOLDER = auto()


class TripleObjectType(Enum):
    """
    Maps to an xsd data type in case of literals, for example: BOOLEAN -> xsd:boolean
    Or Indicates that it is an IRI, in which case it is not an RDF Literal
    """

    BOOLEAN = auto()
    DATETIME = auto()
    DECIMAL = auto()
    INTEGER = auto()
    IRI = auto()
    STRING = auto()
    URI = auto()



import importlib.resources

from rdflib import RDF
from rdflib import SH
from rdflib import Graph

from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.models.input_problems import OntologyResourceProblem
from dsp_tools.commands.validate_data.models.input_problems import OntologyValidationProblem
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias

LIST_SEPARATOR = "\n    - "


def validate_ontology(
    onto_graph: Graph, shacl_validator: ShaclValidator, config: ValidateDataConfig
) -> OntologyValidationProblem | None:
    """
    The API accepts erroneous cardinalities in the ontology.
    To distinguish a mistake in the data from the erroneous ontology, the ontology will be validated beforehand.
    This way, we do not have to take an erroneous ontology into account when validating the data.

    Args:
        onto_graph: the graph of the project ontologies
        shacl_validator: connection to the API for the validation
        config: The configuration where to save the information to

    Returns:
        A validation report if errors were found
    """
    shacl_file = importlib.resources.files("dsp_tools").joinpath("resources/validate_data/validate-ontology.ttl")
    onto_shacl = Graph()
    onto_shacl = onto_shacl.parse(str(shacl_file))
    validation_result = shacl_validator.validate_ontology(onto_graph, onto_shacl)
    if validation_result.conforms:
        return None
    if config.save_graph_dir:
        validation_result.validation_graph.serialize(f"{config.save_graph_dir}_ONTO_VIOLATIONS.ttl")
    return OntologyValidationProblem(_reformat_ontology_validation_result(validation_result.validation_graph))


def _reformat_ontology_validation_result(validation_result: Graph) -> list[OntologyResourceProblem]:
    bns = validation_result.subjects(RDF.type, SH.ValidationResult)
    return [_get_one_problem(validation_result, bn) for bn in bns]


def _get_one_problem(val_g: Graph, result_bn: SubjectObjectTypeAlias) -> OntologyResourceProblem:
    iri = next(val_g.objects(result_bn, SH.focusNode))
    iri_str = reformat_onto_iri(iri)
    msg = str(next(val_g.objects(result_bn, SH.resultMessage)))
    splt = [x.strip() for x in msg.split("\n")]
    return OntologyResourceProblem(iri_str, " ".join(splt))



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.validate_data import validate_data
from test.e2e.setup_testcontainers.ports import ExternalContainerPorts


# ruff: noqa: ARG001 Unused function argument
@pytest.fixture(scope="module")
def create_project_systematic(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/json-project/test-project-systematic.json"), creds)


@pytest.fixture(scope="module")
def api_url(container_ports: ExternalContainerPorts) -> str:
    return f"http://0.0.0.0:{container_ports.api}"


@pytest.fixture(scope="module")
def shacl_validator(api_url: str) -> ShaclValidator:
    return ShaclValidator(api_url)


@pytest.mark.usefixtures("create_project_systematic")
def test_systematic(api_url: str, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/xml-data/test-data-systematic.xml")
    creds = ServerCredentials("root@example.com", "test", api_url)
    no_violations = validate_data(file, creds, save_graphs=False)
    assert no_violations



from __future__ import annotations

import pandas as pd
import pytest


@pytest.fixture
def prefixes_good() -> pd.DataFrame:
    return pd.DataFrame(
        {"prefixes": ["foaf:", "sdh"], "uri": ["http://xmlns.com/foaf/0.1/", "https://ontome.net/ns/sdhss/"]}
    )


@pytest.fixture
def prefixes_missing_col() -> pd.DataFrame:
    return pd.DataFrame({"uri": ["http://xmlns.com/foaf/0.1/", "https://ontome.net/ns/sdhss/"]})


@pytest.fixture
def prefixes_missing_val() -> pd.DataFrame:
    return pd.DataFrame(
        {"prefixes": ["foaf:", pd.NA], "uri": ["http://xmlns.com/foaf/0.1/", "https://ontome.net/ns/sdhss/"]}
    )


@pytest.fixture
def prefixes_wrong_val() -> pd.DataFrame:
    return pd.DataFrame({"prefixes": ["foaf:", "sdh"], "uri": ["http://xmlns.com/foaf/0.1/", "not a uri"]})


@pytest.fixture
def licenses_good() -> pd.DataFrame:
    return pd.DataFrame({"enabled": ["http://rdfh.ch/licenses/cc-by-4.0", "http://rdfh.ch/licenses/ai-generated"]})


@pytest.fixture
def project_good_missing_zero() -> pd.DataFrame:
    return pd.DataFrame({"shortcode": [11], "shortname": ["name"], "longname": ["long"]})


@pytest.fixture
def project_good_no_zero() -> pd.DataFrame:
    return pd.DataFrame({"shortcode": [1111], "shortname": ["name"], "longname": ["long"]})


@pytest.fixture
def project_missing_col() -> pd.DataFrame:
    return pd.DataFrame({"shortname": ["name"], "longname": ["long"]})


@pytest.fixture
def project_missing_val() -> pd.DataFrame:
    return pd.DataFrame({"shortcode": [pd.NA], "shortname": ["name"], "longname": ["long"]})


@pytest.fixture
def project_too_many_rows() -> pd.DataFrame:
    return pd.DataFrame({"shortcode": [11, 0], "shortname": ["name", pd.NA], "longname": ["long", "other"]})


@pytest.fixture
def description_good() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "description_en": ["english"],
            "description_de": [pd.NA],
            "fr": ["french"],
            "description_rm": [pd.NA],
        }
    )


@pytest.fixture
def description_too_many_rows() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "description_en": ["english", "en1"],
            "description_de": [pd.NA, pd.NA],
            "fr": ["french", "french2"],
            "description_it": [pd.NA, "it"],
            "description_rm": [pd.NA, pd.NA],
        }
    )


@pytest.fixture
def description_missing_col() -> pd.DataFrame:
    return pd.DataFrame({"other": ["english"]})


@pytest.fixture
def description_missing_val() -> pd.DataFrame:
    return pd.DataFrame({"other": ["english"], "description_en": [pd.NA]})


@pytest.fixture
def keywords_good() -> pd.DataFrame:
    return pd.DataFrame({"keywords": ["one", pd.NA, "three"]})


@pytest.fixture
def keywords_missing_col() -> pd.DataFrame:
    return pd.DataFrame({"other": [1]})


@pytest.fixture
def keywords_missing_val() -> pd.DataFrame:
    return pd.DataFrame({"other": [1], "keywords": [pd.NA]})


@pytest.fixture
def users_good() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "username": ["Alice", "Caterpillar", "WhiteRabbit"],
            "email": ["alice@dasch.swiss", "caterpillar@dasch.swiss", "white.rabbit@dasch.swiss"],
            "givenname": ["Alice Pleasance", "Caterpillar", "White"],
            "familyname": ["Liddell", "Wonderland", "Rabbit"],
            "password": ["alice4322", "alice7652", "alice8711"],
            "lang": ["en", "de", "fr"],
            "role": ["systemadmin", "projectadmin", "projectmember"],
        }
    )


@pytest.fixture
def users_missing_col() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "username": ["Alice"],
            "email": ["alice@dasch.swiss"],
            "givenname": ["Alice Pleasance"],
            "familyname": ["Liddell"],
            "lang": ["fr"],
            "role": ["systemadmin"],
        }
    )


@pytest.fixture
def users_missing_val() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "username": [pd.NA],
            "email": ["alice@dasch.swiss"],
            "givenname": ["Alice Pleasance"],
            "familyname": ["Liddell"],
            "password": ["alice7652"],
            "lang": ["fr"],
            "role": ["systemadmin"],
        }
    )


@pytest.fixture
def users_wrong_lang() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "username": ["Alice"],
            "email": ["alice@dasch.swiss"],
            "givenname": ["Alice Pleasance"],
            "familyname": ["Liddell"],
            "password": ["alice7652"],
            "lang": ["other"],
            "role": ["systemadmin"],
        }
    )


@pytest.fixture
def user_good() -> pd.Series[str]:
    return pd.Series(
        {
            "username": "Alice",
            "email": "alice@dasch.swiss",
            "givenname": "Alice Pleasance",
            "familyname": "Liddell",
            "password": "alice7652",
            "lang": "en",
            "role": "systemadmin",
        }
    )


@pytest.fixture
def user_wrong_lang() -> pd.Series[str]:
    return pd.Series(
        {
            "username": "Alice",
            "email": "alice@dasch.swiss",
            "givenname": "Alice Pleasance",
            "familyname": "Liddell",
            "password": "alice7652",
            "lang": "other",
            "role": "systemadmin",
        }
    )


@pytest.fixture
def user_wrong_email() -> pd.Series[str]:
    return pd.Series(
        {
            "username": "Alice",
            "email": "not an email",
            "givenname": "Alice Pleasance",
            "familyname": "Liddell",
            "password": "alice7652",
            "lang": "fr",
            "role": "systemadmin",
        }
    )


@pytest.fixture
def user_wrong_role() -> pd.Series[str]:
    return pd.Series(
        {
            "username": "Alice",
            "email": "alice@dasch.swiss",
            "givenname": "Alice Pleasance",
            "familyname": "Liddell",
            "password": "alice7652",
            "lang": "fr",
            "role": "other",
        }
    )



import importlib.resources
from datetime import datetime
from pathlib import Path

from loguru import logger
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.clients.legal_info_client_live import LegalInfoClientLive
from dsp_tools.commands.validate_data.api_clients import ListClient
from dsp_tools.commands.validate_data.api_clients import OntologyClient
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.get_rdf_like_data import get_rdf_like_data
from dsp_tools.commands.validate_data.get_user_validation_message import get_user_message
from dsp_tools.commands.validate_data.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.make_data_graph import make_data_graph
from dsp_tools.commands.validate_data.models.api_responses import EnabledLicenseIris
from dsp_tools.commands.validate_data.models.api_responses import ListLookup
from dsp_tools.commands.validate_data.models.api_responses import OneList
from dsp_tools.commands.validate_data.models.api_responses import ProjectDataFromApi
from dsp_tools.commands.validate_data.models.input_problems import OntologyResourceProblem
from dsp_tools.commands.validate_data.models.input_problems import OntologyValidationProblem
from dsp_tools.commands.validate_data.models.input_problems import SortedProblems
from dsp_tools.commands.validate_data.models.input_problems import UnknownClassesInData
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import RDFGraphStrings
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.sparql.construct_shacl import construct_shapes_graphs
from dsp_tools.commands.validate_data.utils import reformat_onto_iri
from dsp_tools.commands.validate_data.validate_ontology import validate_ontology
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_CYAN
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_GREEN
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_RED
from dsp_tools.utils.ansi_colors import BACKGROUND_BOLD_YELLOW
from dsp_tools.utils.ansi_colors import BOLD_CYAN
from dsp_tools.utils.ansi_colors import BOLD_RED
from dsp_tools.utils.ansi_colors import BOLD_YELLOW
from dsp_tools.utils.ansi_colors import RESET_TO_DEFAULT
from dsp_tools.utils.data_formats.uri_util import is_prod_like_server
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.get_lookups import get_authorship_lookup
from dsp_tools.utils.xml_parsing.get_parsed_resources import get_parsed_resources
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import parse_and_clean_xml_file

LIST_SEPARATOR = "\n    - "


VALIDATION_ERRORS_FOUND_MSG = BACKGROUND_BOLD_RED + "\n   Validation errors found!   " + RESET_TO_DEFAULT
NO_VALIDATION_ERRORS_FOUND_MSG = BACKGROUND_BOLD_GREEN + "\n   No validation errors found!   " + RESET_TO_DEFAULT


def validate_data(filepath: Path, creds: ServerCredentials, save_graphs: bool) -> bool:
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        save_graphs: if this flag is set, all the graphs will be saved in a folder
        creds: server credentials for authentication

    Returns:
        True if no errors that impede an xmlupload were found.
        Warnings and user info do not impede an xmlupload.
    """
    graph_save_dir = None
    if save_graphs:
        graph_save_dir = _get_graph_save_dir(filepath)
    config = ValidateDataConfig(filepath, graph_save_dir, ValidationSeverity.INFO, is_prod_like_server(creds.server))
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    graphs, used_iris = _prepare_data_for_validation_from_file(filepath, auth)
    return _validate_data(graphs, used_iris, auth, config)


def validate_parsed_resources(
    parsed_resources: list[ParsedResource],
    authorship_lookup: dict[str, list[str]],
    permission_ids: list[str],
    shortcode: str,
    config: ValidateDataConfig,
    auth: AuthenticationClient,
) -> bool:
    rdf_graphs, used_iris = _prepare_data_for_validation_from_parsed_resource(
        parsed_resources, authorship_lookup, permission_ids, auth, shortcode
    )
    return _validate_data(rdf_graphs, used_iris, auth, config)


def _validate_data(
    graphs: RDFGraphs, used_iris: set[str], auth: AuthenticationClient, config: ValidateDataConfig
) -> bool:
    if unknown_classes := _check_for_unknown_resource_classes(graphs, used_iris):
        msg = _get_msg_str_unknown_classes_in_data(unknown_classes)
        logger.error(msg)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(msg)
        # if unknown classes are found, we cannot validate all the data in the file
        return False
    shacl_validator = ShaclValidator(auth.server)
    onto_validation_result = validate_ontology(graphs.ontos, shacl_validator, config)
    if onto_validation_result:
        msg = _get_msg_str_ontology_validation_violation(onto_validation_result)
        logger.error(msg)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(msg)
        # if the ontology itself has errors, we will not validate the data
        return False
    report = _get_validation_result(graphs, shacl_validator, config)
    if report.conforms:
        logger.debug("No validation errors found.")
        print(NO_VALIDATION_ERRORS_FOUND_MSG)
        return True
    reformatted = reformat_validation_graph(report)
    sorted_problems = sort_user_problems(reformatted)
    _print_shacl_validation_violation_message(sorted_problems, report, config)
    return _get_validation_status(sorted_problems, config.is_on_prod_server)


def _get_validation_status(all_problems: SortedProblems, is_on_prod: bool) -> bool:
    violations = any(
        [
            bool(all_problems.unique_violations),
            bool(all_problems.unexpected_shacl_validation_components),
        ]
    )
    if violations:
        return False
    if is_on_prod and all_problems.user_warnings:
        return False
    return True


def _prepare_data_for_validation_from_file(filepath: Path, auth: AuthenticationClient) -> tuple[RDFGraphs, set[str]]:
    parsed_resources, shortcode, authorship_lookup, permission_ids = _get_info_from_xml(filepath, auth.server)
    return _prepare_data_for_validation_from_parsed_resource(
        parsed_resources, authorship_lookup, permission_ids, auth, shortcode
    )


def _get_info_from_xml(file: Path, api_url: str) -> tuple[list[ParsedResource], str, dict[str, list[str]], list[str]]:
    root = parse_and_clean_xml_file(file)
    shortcode = root.attrib["shortcode"]
    authorship_lookup = get_authorship_lookup(root)
    permission_ids = [perm.attrib["id"] for perm in root.findall("permissions")]
    parsed_resources = get_parsed_resources(root, api_url)
    return parsed_resources, shortcode, authorship_lookup, permission_ids


def _prepare_data_for_validation_from_parsed_resource(
    parsed_resources: list[ParsedResource],
    authorship_lookup: dict[str, list[str]],
    permission_ids: list[str],
    auth: AuthenticationClient,
    shortcode: str,
) -> tuple[RDFGraphs, set[str]]:
    used_iris = {x.res_type for x in parsed_resources}
    proj_info = _get_project_specific_information_from_api(auth, shortcode)
    list_lookup = _make_list_lookup(proj_info.all_lists)
    data_rdf = _make_data_graph_from_parsed_resources(parsed_resources, authorship_lookup, list_lookup)
    rdf_graphs = _create_graphs(data_rdf, shortcode, auth, proj_info, permission_ids)
    return rdf_graphs, used_iris


def _make_list_lookup(project_lists: list[OneList]) -> ListLookup:
    lookup = {}
    for li in project_lists:
        for nd in li.nodes:
            lookup[(li.list_name, nd.name)] = nd.iri
            lookup[("", nd.iri)] = nd.iri
    return ListLookup(lookup)


def _get_project_specific_information_from_api(auth: AuthenticationClient, shortcode: str) -> ProjectDataFromApi:
    list_client = ListClient(auth.server, shortcode)
    all_lists = list_client.get_lists()
    enabled_licenses = _get_license_iris(shortcode, auth)
    return ProjectDataFromApi(all_lists, enabled_licenses)


def _make_data_graph_from_parsed_resources(
    parsed_resources: list[ParsedResource], authorship_lookup: dict[str, list[str]], list_lookup: ListLookup
) -> Graph:
    rdf_like_data = get_rdf_like_data(parsed_resources, authorship_lookup, list_lookup)
    rdf_data = make_data_graph(rdf_like_data)
    return rdf_data


def _get_msg_str_unknown_classes_in_data(unknown: UnknownClassesInData) -> str:
    if unknown_onto_msg := _get_unknown_ontos_msg(unknown):
        return unknown_onto_msg
    unknown_classes = sorted(list(unknown.unknown_classes))
    known_classes = sorted(list(unknown.defined_classes))
    return (
        f"Your data uses resource classes that do not exist in the ontologies in the database.\n"
        f"The following classes that are used in the data are unknown: {', '.join(unknown_classes)}\n"
        f"The following classes exist in the uploaded ontologies: {', '.join(known_classes)}"
    )


def _get_unknown_ontos_msg(unknown: UnknownClassesInData) -> str | None:
    def split_prefix(relative_iri: str) -> str | None:
        if ":" not in relative_iri:
            return None
        return relative_iri.split(":")[0]

    used_ontos = set(not_knora for x in unknown.unknown_classes if (not_knora := split_prefix(x)))
    exising_ontos = set(not_knora for x in unknown.defined_classes if (not_knora := split_prefix(x)))
    if unknown_found := used_ontos - exising_ontos:
        return (
            f"Your data uses ontologies that don't exist in the database.\n"
            f"The following ontologies that are used in the data are unknown: {', '.join(unknown_found)}\n"
            f"The following ontologies are uploaded: {', '.join(exising_ontos)}"
        )
    return None


def _get_msg_str_ontology_validation_violation(onto_violations: OntologyValidationProblem) -> str:
    probs = sorted(onto_violations.problems, key=lambda x: x.res_iri)

    def get_resource_msg(res: OntologyResourceProblem) -> str:
        return f"Resource Class: {res.res_iri} | Problem: {res.msg}"

    problems = [get_resource_msg(x) for x in probs]
    return (
        "The ontology structure contains errors that prevent the validation of the data.\n"
        "Please correct the following errors and re-upload the corrected ontology.\n"
        f"Once those two steps are done, the command `validate-data` will find any problems in the data.\n"
        f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(problems)}"
    )


def _print_shacl_validation_violation_message(
    sorted_problems: SortedProblems, report: ValidationReportGraphs, config: ValidateDataConfig
) -> None:
    messages = get_user_message(sorted_problems, config.xml_file)
    if messages.violations:
        logger.error(messages.violations.message_header, messages.violations.message_body)
        print(VALIDATION_ERRORS_FOUND_MSG)
        print(BOLD_RED, messages.violations.message_header, RESET_TO_DEFAULT)
        print(messages.violations.message_body)
    else:
        logger.debug("No validation errors found.")
        print(NO_VALIDATION_ERRORS_FOUND_MSG)
    if messages.warnings and config.severity.value <= 2:
        logger.warning(messages.warnings.message_header, messages.warnings.message_body)
        print(BACKGROUND_BOLD_YELLOW + "\n    Warning!    " + RESET_TO_DEFAULT)
        print(BOLD_YELLOW, messages.warnings.message_header, RESET_TO_DEFAULT)
        print(messages.warnings.message_body)
    if messages.infos and config.severity.value == 1:
        logger.info(messages.infos.message_header, messages.infos.message_body)
        print(BACKGROUND_BOLD_CYAN + "\n    Potential Problems Found    " + RESET_TO_DEFAULT)
        print(BOLD_CYAN, messages.infos.message_header, RESET_TO_DEFAULT)
        print(messages.infos.message_body)
    if messages.unexpected_violations:
        logger.error(messages.unexpected_violations.message_header, messages.unexpected_violations.message_body)
        print(
            BACKGROUND_BOLD_RED,
            "\n    Unknown violations found!   ",
            RESET_TO_DEFAULT,
        )
        if config.save_graph_dir:
            print(
                BOLD_RED,
                messages.unexpected_violations.message_header,
                "Consult the saved graphs for details.",
                RESET_TO_DEFAULT,
            )
            print(messages.unexpected_violations.message_body)
        else:
            _save_unexpected_results_and_inform_user(report, config.xml_file)


def _save_unexpected_results_and_inform_user(report: ValidationReportGraphs, filepath: Path) -> None:
    timestamp = f"{datetime.now()!s}_"
    save_path = filepath.parent / f"{timestamp}_validation_result.ttl"
    report.validation_graph.serialize(save_path)
    shacl_p = filepath.parent / f"{timestamp}_shacl.ttl"
    report.shacl_graph.serialize(shacl_p)
    data_p = filepath.parent / f"{timestamp}_data.ttl"
    report.data_graph.serialize(data_p)
    msg = (
        f"\nPlease contact the development team with the files starting with the timestamp '{timestamp}' "
        f"in the directory '{filepath.parent}'."
    )
    print(BOLD_RED + msg + RESET_TO_DEFAULT)


def _check_for_unknown_resource_classes(
    rdf_graphs: RDFGraphs, used_resource_iris: set[str]
) -> UnknownClassesInData | None:
    res_cls = _get_all_onto_classes(rdf_graphs)
    if extra_cls := used_resource_iris - res_cls:
        unknown_classes = {reformat_onto_iri(x) for x in extra_cls}
        defined_classes = {reformat_onto_iri(x) for x in res_cls}
        return UnknownClassesInData(unknown_classes=unknown_classes, defined_classes=defined_classes)
    return None


def _get_all_onto_classes(rdf_graphs: RDFGraphs) -> set[str]:
    ontos = rdf_graphs.ontos + rdf_graphs.knora_api
    is_resource_iri = URIRef(KNORA_API_STR + "isResourceClass")
    resource_classes = set(ontos.subjects(is_resource_iri, Literal(True)))
    is_usable = URIRef(KNORA_API_STR + "canBeInstantiated")
    usable_resource_classes = set(ontos.subjects(is_usable, Literal(True)))
    user_facing = usable_resource_classes.intersection(resource_classes)
    return {str(x) for x in user_facing}


def _get_validation_result(
    rdf_graphs: RDFGraphs, shacl_validator: ShaclValidator, config: ValidateDataConfig
) -> ValidationReportGraphs:
    report = _validate(shacl_validator, rdf_graphs, config.save_graph_dir)
    if config.save_graph_dir:
        report.validation_graph.serialize(f"{config.save_graph_dir}_VALIDATION_REPORT.ttl")
    return report


def _create_graphs(
    data_rdf: Graph,
    shortcode: str,
    auth: AuthenticationClient,
    proj_info: ProjectDataFromApi,
    permission_ids: list[str],
) -> RDFGraphs:
    logger.debug("Create all graphs.")
    onto_client = OntologyClient(auth.server, shortcode)
    ontologies = _get_project_ontos(onto_client)
    knora_ttl = onto_client.get_knora_api()
    knora_api = Graph()
    knora_api.parse(data=knora_ttl, format="ttl")
    shapes = construct_shapes_graphs(ontologies, knora_api, proj_info, permission_ids)
    api_shapes = Graph()
    api_shapes_path = importlib.resources.files("dsp_tools").joinpath("resources/validate_data/api-shapes.ttl")
    api_shapes.parse(str(api_shapes_path))
    api_card_shapes = Graph()
    api_card_path = importlib.resources.files("dsp_tools").joinpath(
        "resources/validate_data/api-shapes-resource-cardinalities.ttl"
    )
    api_card_shapes.parse(str(api_card_path))
    content_shapes = shapes.content + api_shapes
    card_shapes = shapes.cardinality + api_card_shapes
    return RDFGraphs(
        data=data_rdf,
        ontos=ontologies,
        cardinality_shapes=card_shapes,
        content_shapes=content_shapes,
        knora_api=knora_api,
    )


def _get_project_ontos(onto_client: OntologyClient) -> Graph:
    logger.debug("Get project ontologies from server.")
    all_ontos = onto_client.get_ontologies()
    onto_g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        onto_g += og
    return onto_g


def _get_license_iris(shortcode: str, auth: AuthenticationClient) -> EnabledLicenseIris:
    legal_client = LegalInfoClientLive(auth.server, shortcode, auth)
    license_info = legal_client.get_licenses_of_a_project()
    iris = [x["id"] for x in license_info]
    return EnabledLicenseIris(iris)


def _validate(validator: ShaclValidator, rdf_graphs: RDFGraphs, graph_save_dir: Path | None) -> ValidationReportGraphs:
    logger.debug("Serialise RDF graphs into turtle strings")
    data_str = rdf_graphs.data.serialize(format="ttl")
    ontos_str = rdf_graphs.ontos.serialize(format="ttl")
    card_shape_str = rdf_graphs.cardinality_shapes.serialize(format="ttl")
    content_shape_str = rdf_graphs.content_shapes.serialize(format="ttl")
    knora_api_str = rdf_graphs.knora_api.serialize(format="ttl")
    graph_strings = RDFGraphStrings(
        cardinality_validation_data=data_str,
        cardinality_shapes=card_shape_str + ontos_str + knora_api_str,
        content_validation_data=data_str + ontos_str + knora_api_str,
        content_shapes=content_shape_str + ontos_str + knora_api_str,
    )
    if graph_save_dir:
        _save_graphs(graph_save_dir, graph_strings)
    validation_results = validator.validate(graph_strings)
    return ValidationReportGraphs(
        conforms=validation_results.conforms,
        validation_graph=validation_results.validation_graph,
        shacl_graph=rdf_graphs.cardinality_shapes + rdf_graphs.content_shapes,
        onto_graph=rdf_graphs.ontos + rdf_graphs.knora_api,
        data_graph=rdf_graphs.data,
    )


def _save_graphs(save_path: Path, graph_strings: RDFGraphStrings) -> None:
    with open(f"{save_path}_CARDINALITY_DATA.ttl", "w") as writer:
        writer.write(graph_strings.cardinality_validation_data)
    with open(f"{save_path}_CARDINALITY_SHAPES.ttl", "w") as writer:
        writer.write(graph_strings.cardinality_shapes)
    with open(f"{save_path}_CONTENT_DATA.ttl", "w") as writer:
        writer.write(graph_strings.content_validation_data)
    with open(f"{save_path}_CONTENT_SHAPES.ttl", "w") as writer:
        writer.write(graph_strings.content_shapes)


def _get_graph_save_dir(filepath: Path) -> Path:
    parent_directory = filepath.parent
    new_directory = parent_directory / "graphs"
    new_directory.mkdir(exist_ok=True)
    save_file_template = new_directory / filepath.stem
    print(BOLD_CYAN + f"\n   Saving graphs to {save_file_template}   " + RESET_TO_DEFAULT)
    return save_file_template
